---
author:
- Orestis Malaspinas
title: Sciences pour orientation logicielle
autoSectionLabels: false
autoEqnLabels: true
eqnPrefix:
  - "éq."
  - "éqs."
chapters: true
numberSections: false
chaptersDepth: 1
sectionsDepth: 3
lang: fr
documentclass: book
papersize: A4
cref: false
urlcolor: blue
---
\newcommand{\real}{\mathbb{R}}
\newcommand{\rational}{\mathbb{Q}}
\newcommand{\integer}{\mathbb{Z}}
\renewcommand{\natural}{\mathbb{N}}
\newcommand{\complex}{\mathbb{C}}
\newcommand{\zbar}{\bar{z}}
\newcommand{\dd}{\mathrm{d}}
\newcommand{\perm}{\mathrm{perm}}
\newcommand{\card}{\mathrm{card}}
\newcommand{\fh}{\hat{f}}
\newcommand{\gh}{\hat{g}}
\newcommand{\hh}{\hat{h}}
\renewcommand{\Re}{\mathrm{Re}}
\renewcommand{\Im}{\mathrm{Im}}
\newcommand{\pDeriv}[2]{\frac{\partial #1}{\partial #2}}
\newcommand{\pDerivTwo}[2]{\frac{\partial^2 #1}{\partial #2^2}}
\newcommand{\dDeriv}[2]{\frac{\dd #1}{\dd #2}}
\newcommand{\dDerivTwo}[2]{\frac{\dd^2 #1}{\dd #2^2}}
\newcommand{\cm}{\mathrm{cm}}
\newcommand{\km}{\mathrm{km}}
\newcommand{\mm}{\mathrm{mm}}
\newcommand{\cd}{\mathrm{cd}}
\newcommand{\mol}{\mathrm{mol}}
\newcommand{\m}{\mathrm{m}}
\newcommand{\s}{\mathrm{s}}
\newcommand{\kg}{\mathrm{kg}}
\newcommand{\g}{\mathrm{g}}
\newcommand{\K}{\mathrm{K}}
\newcommand{\J}{\mathrm{J}}
\newcommand{\oC}{^\circ\mathrm{C}}
\newcommand{\oK}{^\circ\mathrm{K}}
\newcommand{\A}{\mathrm{A}}
\newcommand{\N}{\mathrm{N}}
\newcommand{\atm}{\mathrm{atm}}
\newcommand{\V}{\mathrm{V}}
\newcommand{\W}{\mathrm{W}}
\newcommand{\kW}{\mathrm{kW}}
\newcommand{\dl}{\mathrm{dl}}
\newcommand{\dm}{\mathrm{dm}}
\newcommand{\kcal}{\mathrm{kcal}}
\newcommand{\h}{\mathrm{h}}
\newcommand{\Pa}{\mathrm{Pa}}
\newcommand{\vectwo}[2]{\begin{pmatrix}#1 \\ #2 \end{pmatrix}}
\newcommand{\vecthree}[3]{\begin{pmatrix}#1 \\ #2 \\ #3 \end{pmatrix}}
\newcommand{\mat}[1]{\underline{\underline{#1}}}
\newcommand{\com}[1]{\mat{\mathrm{com}#1}}
\newcommand{\mattwo}[4]{\begin{pmatrix}
                #1 & #2 \\
                #3 & #4
            \end{pmatrix}}

# Rappel

## Les fonctions

Une fonction, notée $f$, est une relation entre deux ensembles. Soient deux ensembles $X$ et $Y$, cette fonction va associer à chaque élément de $X$ à un élément de $Y$.

On note de façon formelle
$$f:X\rightarrow Y.$$
Ici $X$ est le *domaine de définition* de $f$ et $Y$ est le *domaine d'arrivée*.
Cette notation nous donne juste le domaine de définition et le domaine d'arrivée de $f$ mais ne nous dit pas quelle est la règle d'association entre les éléments de $X$ et les éléments de $Y$.

Pour un exemple concret, on peut prendre
$$f:\real\rightarrow\real,$$
où $f$ est définie par
$$f:x\rightarrow x^2,\quad \mbox{ou}\quad f(x)=x^2.$$

On a donc ici que $\real$ est le domaine de définition et d'arrivée de $f$. En revanche tout $\real$ n'est pas "couvert" par les éléments de $f(x)$.

## Fonctions polynomiales

Une fonction polynomiale de $\real\rightarrow\real$ de degré $n\in\natural$ est une fonction du genre
\begin{align}
&p_n:X\rightarrow Y,\\
&p_n\rightarrow a_0+a_1\cdot x+a_2\cdot x^2+\dots+a_n\cdot x^n,
\end{align}
où $\{a_i\}_{i=0}^n\in\real$ sont des coefficients réels.

## Les dérivées

Lorsque nous étudions les fonction il est fondamental d'étudier leur taux de variation.
Soit $f$, une fonction réelle, définie sur un intervalle $[a,b]\subseteq\real$
$$
f:[a,b]\rightarrow\real.
$$
Soit $x_0\in]a,b[$, nous cherchons à connaître le taux de variation de $f$ en ce point.
Pour ce faire nous allons prendre un point dans le voisinage de $x_0$, $x_1=x_0+h\in]a,b[$, et $h$ "petit".
Nous pouvons dès lors calculer le taux de variation de $f$, également appelé **différences divisées**
$$
\frac{f(x_0+h)-f(x_0)}{h}=\frac{f(x_1)-f(x_0)}{x_1-x_0}.
$${#eq:diff_div}
Plus $h$ sera petit, plus cette approximation sera bonne. En prenant la *limite* pour
$h\rightarrow 0$, on obtient la **dérivée** de $f$ en $x_0$, notée $f'(x_0)$
$$
f'(x_0)=\lim_{h\rightarrow}\frac{f(x_0+h)-f(x_0)}{h}=\lim_{h\rightarrow}\frac{f(x_1)-f(x_0)}{x_1-x_0}.
$${eq:def_deriv}
Géométriquement la dérivée n'est autre que la pente de la tangente de la fonction $f$ en $x_0$.
En évaluant cette limite pour tous les $x\in]a,b[$, on obtient la dérivée de $f$, nommée $f$ sur tout l'intervalle $]a,b[$.

![La fonction $f(x)=x^3$ (ligne noire), sa dérivée $f'(x)=3\cdot x^2$ (ligne bleue), et deux approximations de
la dérivée $h=0.2$ (croix rouges) et $h=0.4$ (croix vertes).](figs/deriv_approx.svg){#fig:deriv_approx width=70%}

Sur la @fig:deriv_approx, nous voyons représentées la fonction $x^3$ (ligne noire), sa dérivée exacte $3x^2$
et deux approximations avec avec $h=0.2$ (croix rouges) et $h=0.4$ (croix vertes).
On constate que les approximations deviennent meilleures lorsque $h$ devient plus petit, en effet la distance entre
la dérivée exacte de la fonction et ses approximations devient plus petite.

La @fig:deriv_approx est obtenu avec le listing suivant

```{.python}
import numpy as np
import matplotlib.pyplot as plt

def f(x):
    return x**3

def df(x):
    return 3*x**2

def approx_df(x0,x1):
    return (f(x1)-f(x0))/(x1-x0)

dx = 0.1
xmin = -2
xmax = 2 + dx
x = np.arange(xmin,xmax,dx)
x1 = x+dx*2
x2 = x+dx*4
plt.plot(x,f(x),'k-', label="$f(x)=x^3$")
plt.plot(x,df(x),'b-', label="$f'(x)=3\cdot x^2$")
plt.plot(x,approx_df(x,x1),'rx', label="$h=0.2$") # approx 0.2
plt.plot(x,approx_df(x,x2),'gx', label="$h=0.4$") # approx 0.4
plt.legend(loc=4)

plt.grid(True)
plt.show() # show on screen
```

## La série de Taylor

Lorsque nous souhaitons approximer une fonction $f$ réelle, dérivable une infinité de fois au voisinage d'un point, $x_0\in\real$,
nous pouvons utiliser son développement en **série de Taylor** jusqu'à l'ordre $N\in\natural$, qui s'exprime comme
$$
f(x)\cong p_N(x)=\sum_{n=0}^N \frac{1}{n!}(x-x_0)^nf^{(n)}(x_0).
$$
Nous pouvons également évaluer l'erreur
que nous allons commettre lors de cette approximation
$$
E(x)=f(x)-p_N(x).
$$
Il est très difficile de connaître exactement cette erreur, nous savons seulement que pour un certain $\xi\in[x_0,x]$, nous aurons
$$
E(x)=\frac{f^{(N+1)}(\xi)}{(N+1)!}\cdot(x-x_0)^{N+1}.
$$
En d'autres termes l'erreur est proportionnelle à la dérivée $N+1$ en un point donné de $[x_0,x]$,
et à la $(N+1)$-ème puissance de $(x-x_0)$ (si $x$ est proche de $x_0$ cette différence tend rapidement vers $0$). Pour un $N$ fixé, il suffit de poser $\mathcal{K}=\frac{f^{(N+1)}(\xi)}{(N+1)!}\in\real$, et on va avoir
que l'erreur est donnée par
$$
E(x)=\mathcal{K}\cdot(x-x_0)^{N+1}.
$${#eq:err_taylor}
A l'aide de cette expression, on voit bien que l'erreur tend vers $0$ si $x$ tend vers $x_0$.

---

Exemple +.#

Essayons de développer $f(x)=e^x$ en série de Taylor autour d'un point $x_0$ quelconque.
On doit d'abord calculer ses dérivées. Il se trouve que
$$
f'(x)=(e^x)'=e^x.
$$
On a donc aisément que la série de Taylor de $e^x$ pour n'importe quel $x_0$ et un $N\in\natural$ est donnée par
$$
p_N(x)=\sum_{n=0}^N\frac{e^{x_0}}{n!}(x-x_0)^n.
$$
Si $x_0=0$, il vient
$$
p_N(x)=\sum_{n=0}^N\frac{1}{n!}x^n.
$$
On peut à présent mesurer l'erreur que nous effectuons. Pour un $N$ donné
on va avoir que l'erreur est donnée par (voir l'@eq:err_taylor)
$$
E(x)=e^x-(1+x)=\mathcal{K}x^{N+1}.\nonumber
$$
Fixons $N=1$. On a donc que
$$
p_1(x)=1+x,\mbox{ et } E(x)=\mathcal{K}\cdot x^2.
$$
Si nous à présent nous mesurons $E$ en fonction de $x^2$, on devrait obtenir
un droite de pente $\mathcal{K}$. On peut voir l'approximation (image de gauche), ainsi que l'erreur
de l'approximation (image de droite) sur la @fig:taylor_exp. On constate bien que l'erreur est proportionnelle à $x^2$ (elle est donc croissante avec $x$).

![On voit la fonction exponentielle (en noir), ainsi que son approximation pour $N=1$ (en rouge) sur l'image de gauche. Sur l'image de droite on voit l'erreur de l'approximation en fonction de $x^2$, ainsi qu'une droite dont la pente est 0.525.](figs/taylor_exp.svg){#fig:taylor_exp width=70%}

L'image @fig:taylor_exp a été obtenue avec le listing suivant

```python
import numpy as np
import matplotlib.pyplot as plt
import math

def f(x):
    return np.exp(x)

def p(x):
    return 1+x

def df(x):
    return np.exp(x)

x0 = 0
dx = np.arange(0,0.3,0.05)
x = np.array([])
e = np.array([])
for delta in dx:
    x = np.append(x, x0 + delta)
    e = np.append(e, f(x0 + delta)-p(x0 + delta))

plt.figure(1)

plt.subplot(121)
plt.title('Exponentielle')
plt.plot(x, f(x), 'kx-', label="$e^x$")
plt.plot(x, p(x), 'rx-',  label="$1+x$")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.title('Erreur')
plt.plot(x**2, e,   'kx-', label="$E(x)$")
plt.plot(np.arange(0,0.3**2,0.05), 0.525*np.arange(0,0.3**2,0.05), 'r-',  label="$0.525\cdot x$")
plt.xlabel('$x^2$')
plt.legend(loc=0)
plt.grid(True)
plt.show()
```

---

Pour résumer lorsque nous faisons la série de Taylor d'une fonction jusqu'à un ordre $N$, l'erreur
sera proportionnelle à $(x-x_0)^{N+1}$, qu'on écrit souvent comme $\mathcal{O}((x-x_0))^{N+1}$
$$
f(x)=\sum_{n=0}^N\frac{1}{n!}f^{(n)}(x_0)(x-x_0)^n+\mathcal{O}((x-x_0)^{N+1}).
$${#eq:taylor_err}
On dit que **l'ordre** de l'approximation est de $N+1$ (l'erreur va décroître comme $(x-x_0)^{N+1}$).

# L'approximation des dérivées

De ce que nous venons de voir dans la section précédente, il est possible d'approximer les dérivées de fonctions à l'aide de calculs assez simples.
Par ailleurs, ces estimations numériques permettent également de travailler sur des échantillons de fonctions inconnues (dans le cas du traitement d'un
signal par exemple, ou la pixellisation d'une image).

![Un signal échantillonné (en noir) ainsi qu'une approximation de sa dérivée (en rouge).](figs/sin_noise.svg){#fig:sin_noise width=70%}

Sur la @fig:sin_noise, on voit un signal échantillonné (en noir) et sa dérivée (en rouge) calculée numériquement à l'aide de la formule de la différence divisée
(voir l'@eq:diff_div). Nous allons voir dans ce chapitre différentes façon d'approximer une dérivée
numériquement. Nous étudieront également l'erreur effectuée lors de ces approximations.

Afin de créer la @fig:sin_noise nous avons utilisé le listing suivant

```{.python}
import numpy as np
import matplotlib.pyplot as plt
import math

def f(x): # sinus + bruit
    return np.sin(x) + np.random.rand(len(x))*0.1

def approx_df(x,dx): # forward finite diff
    tmp = f(x)
    return (np.roll(tmp,-1)-tmp) / dx

dx = 0.2
xmin = -math.pi
xmax = math.pi-dx
x = np.arange(xmin,xmax,dx) # interval
np.random.seed(0) # init noise
plt.plot(x,f(x),'k-o', label="$f(x)$")

np.random.seed(0) # init niose
plt.plot(x,approx_df(x,dx),'r-o', label="$f'(x), approx$")
plt.legend(loc=2)

plt.grid(True)
plt.show() # plot on the screen
```

## Les dérivées numériques (différences divisées)

Comme nous l'avons vu dans le chapitre précédent, nous pouvons approximer les dérivées
en mesurant le taux de variation d'une fonction sur une distance $h>0$ "petite". Nous avons
écrit l'@eq:diff_div
$$
f'(x_0)\cong \frac{f(x_0+h)-f(x_0)}{h},
$$
qui s'appelle également formule de **différences finies vers l'avant d'ordre 1**
(nous reviendront sur l'ordre un plus tard).
Nous avons utilisé cette formule en nous inspirant de la définition de la dérivée qui peut s'écrire
à l'aide d'une limite
$$
f'(x_0)=\lim_{h\rightarrow 0} \frac{f(x_0+h)-f(x_0)}{h}.\nonumber
$$
Si cette limite existe, on peut également l'écrire de deux façon différentes
\begin{align}
f'(x_0)&=\lim_{h\rightarrow 0} \frac{f(x_0+h)-f(x_0)}{h},\nonumber\\
&=\lim_{h\rightarrow 0} \frac{f(x_0)-f(x_0-h)}{h},\\
&=\lim_{h\rightarrow 0} \frac{f(x_0+h/2)-f(x_0-h/2)}{h}.
\end{align}
En nous inspirant de ces deux dernières formules, nous pouvons écrire deux nouvelles formules de différences finies: la **différence finie en arrière d'ordre 1**
$$
f'(x_0)\cong \frac{f(x_0)-f(x_0-h)}{h},
$$
et la **différence finie centrale d'ordre 2**
$$
f'(x_0)\cong \frac{f(x_0+h/2)-f(x_0-h/2)}{h}.
$$
Pour chacune de ces méthodes vous pouvez observer l'approximation de la dérivée de
la fonction $f(x)=x^3$ ($f'(x)=3x^2$) sur la @fig:finite_diff. On peut constater que la méthode
centrale d'ordre 2 est bien plus précises que les deux autres. Cela est dû à l'ordre de chacune de ces méthodes: alors que la méthode centrale est d'ordre 2, celles en avant et en arrière sont d'ordre 1 uniquement.
Comme pour le cas du développement de Taylor l'ordre est en fait représentatif
de la précision de l'approximation. Dans la section suivante nous allons justement discuter des erreurs.

![Comparaison entre les différences finies en avant (en croix rouges) et en arrière (ronds verts) d'ordre un et centrées d'ordre deux (carré cyan) avec la dérivée exacte $f'(x)=3x^2$ (en trait plein bleu) de la fonction $f(x)=x^3$ (en trait plein noir).](figs/finite_diff.svg){#fig:finite_diff width=70%}

La @fig:finite_diff est obtenue avec le listing suivant

```python
import numpy as np
import matplotlib.pyplot as plt
import math

def f(x):
    return x**3

def df(x):
    return 3*x**2

dx = 0.2
xmin = 0
xmax = math.pi-dx
x = np.arange(xmin,xmax,dx)
x12 = x + dx/2
x1 = x + dx
x_1 = x - dx

plt.plot(x, f(x), 'k-', label="$f(x)=x^3$")
plt.plot(x, df(x),'b-', label="$f'(x)=2\cdot x^2$")
plt.plot(x,(f(x1)-f(x)) / dx,  'rx', label="$f'(x), avant$")
plt.plot(x,(f(x)-f(x_1)) / dx, 'go', label="$f'(x), arriere$")
plt.plot(x12,(f(x1)-f(x)) / dx,'cs', label="$f'(x), centrale$")

plt.legend(loc=2)

plt.grid(True)
plt.show()
```

## Évaluation de l'erreur

Lorsqu'on effectue des approximations de fonctions,
il est nécessaire d'avoir une idée de la valeur de l'erreur
qu'on fait. Le but de cette sous section est d'introduire une méthode
d'étude de l'erreur d'une approximation.

Il est très difficile en général de connaître exactement l'erreur de l'approximation
pour un $h$ donné. En revanche, on peut connaître l'ordre de l'erreur, c'est-à-dire
comment l'erreur va dépendre de $h$. Pour ce faire nous allons utiliser le développement
de Taylor et ce que nous connaissons de l'erreur de l'approximation en série de Taylor.

### Méthode de différences finies vers l'avant d'ordre 1

Dans cette sous section nous essayons de déterminer l'erreur que nous commettons lorsque
nous calculons la dérivée d'une fonction à l'aide des différences finies vers l'avant d'ordre un.
On veut en fait calculer
$$
E(x)=\left| f'(x)-\frac{f(x+h)-f(x)}{h}\right|,
$${#eq:err_avant}
où $h>0$ et est "petit".
Le développement en série de Taylor de $f(x+h)$ est donné par
$$
f(x+h)=f(x)+hf'(x)+\frac{h^2}{2}f''(\xi),
$${#eq:taylor_un}
où $\xi\in[x,x+h]$.
En substituant cette relation dans l'@eq:err_avant, on obtient
\begin{align}
E(x)&=\left| f'(x)-\frac{f(x)+hf'(x)+\frac{h^2}{2}f''(\xi)-f(x)}{h}\right|,\nonumber\\
&=\left| f'(x)-f'(x)+\frac{hf''(\xi)}{2}\right|=\frac{h}{2}\left|f''(\xi)\right|\leq \mathcal{K}\cdot h,
\end{align}
où $\mathcal{K}\in\real$ est une constante.
On constate donc que l'erreur est d'ordre un, soit $\mathcal{O}(h)$ (proportionnelle à $h$).

Cela veut dire que lorsque nous divisons $h$ par deux, l'erreur va environ décroître d'un facteur 2.
A l'inverse si $h$ est multiplié par deux l'erreur sera également multipliée par un facteur 2.
Néanmoins, il faut également ce rendre compte que le facteur $h$ est multiplié par un nombre, $f''(\xi)$, $\xi\in[x,x+h]$.
Lorsque $h$ va varier, ce nombre variera aussi potentiellement. On aura donc que cette évaluation de l'erreur
sera **bornée** par $\mathcal{K}\cdot h$, où $\mathcal{K}=\max_{h} f''(\xi),\forall h$.

Une illustration de ce comportement peut se voir sur la @fig:deriv_avant_1. On peut y voir l'évolution du calcul de l'approximation
de la dérivée de la fonction $f(x)=x^3$ ($f'(x)=3x^2$). On constate que l'erreur n'est pas exactement une droite, mais qu'elle est bornée
par une droite dont la pente est la valeur maximale de $f''(\xi)$.

![Calcul de la dérivée de $f(x)=x^3$ (à gauche) ainsi que de l'erreur (à droite). On peut constater
que l'erreur est bornée par une droite donc la pente est donnée par le maximum des deuxièmes dérivées
de $f(x)$ sur l'intervalle $[0,0.5]$.](figs/deriv_avant_1.svg){#fig:deriv_avant_1 width=70%}

La @fig:deriv_avant_1 est obtenue à l'aide du listing suivant

```python
import numpy as np
import matplotlib.pyplot as plt

def f(x):
    return x**3

def df(x):
    return 3 * x**2

x0 = 0.0
dx = np.arange(0.05,0.5,0.05)
x = np.array([])
deriv = np.array([])
e = np.array([])
for delta in dx:
    x = np.append(x, x0 + delta)
    dd = (f(x0 + delta)-f(x0))/delta
    deriv = np.append(deriv, dd)
    e = np.append(e, abs(df(x0) - dd))

max_d2 = np.max(e/dx) # calcul du max des deuxièmes dérivées

plt.figure(1)

plt.subplot(121)
plt.title('Fonction')
plt.plot(x, f(x), 'b-', label="$x^3$")
plt.plot(x, df(x), 'kx-', label="$3\cdot x^2$")
plt.plot(x, deriv, 'rx-',  label="$fin. diff. avant$")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.title('Erreur')
plt.plot(dx, e,   'kx-', label="$E(x)$")
plt.plot(dx, dx * max_d2,  'r-', label="$\max_y\{ f''(y)\} \cdot x$")
plt.xlabel('$h$')

plt.legend(loc=0)
plt.grid(True)
plt.show()
```

---

Exercice (Méthode de différences finies vers l'arrière d'ordre 1) +.#

En vous inspirant de ce que nous venons de voir, montrez que
la méthode de différences finies vers l'arrière est bien d'ordre 1.

---

A présent, on peut également évaluer l'erreur totale que nous commettons lorsque nous évaluons la dérivée
avec une approximation. Lorsque nous évaluons la dérivée sur un ensemble de $q$ points $\{x_i\}_{i=0}^{q-1}$, avec
$x_i=x_0+i\cdot h$, on a une fonction d'erreur sur chaque point qui est donnée par
$$
E(x_i)=\left(f'(x_i)-\frac{f(x_i+h)-f(x_i)}{h}\right)^2.
$$
Cette définition est un peu différente de celle vue précédemment où nous utilisons une valeur absolue.
Cela est dû au fait qu'il est beaucoup plus aisé de manipuler des carrés que la valeur absolue (en particulier
sa dérivée est continue contrairement à la valeur aboslue) mais les deux méthodes sont une bonne indication de l'erreur d'une méthode numérique.
On peut donc calculer l'erreur totale, $E_\mathrm{tot}$ en sommant l'erreur sur chaque point,
puis en prenant la racine carrée du tout
$$
E_\mathrm{tot}=\sqrt{\sum_{i=0}^{q-1}E(x_i)}=\sqrt{\sum_{i=0}^{q-1}\left(f'(x_i)-\frac{f(x_i+h)-f(x_i)}{h}\right)^2}.
$${#eq:err_tot}
De façon similaire on peut calculer l'erreur moyenne
$$
E_\mathrm{moy}=\sqrt{\frac{1}{q}\sum_{i=0}^{q-1}E(x_i)}=\sqrt{\frac{1}{q}\sum_{i=0}^{q-1}\left(f'(x_i)-\frac{f(x_i+h)-f(x_i)}{h}\right)^2}.
$${#eq:err_moy}

A l'aide de l'@eq:taylor_un on peut calculer $E_\mathrm{tot}$ et $E_\mathrm{moy}$
\begin{align}
E_\mathrm{tot}=\sqrt{\sum_{i=0}^{q-1}\left(f'(x_i)-\frac{f(x_i+h)-f(x_i)}{h}\right)^2}=\sqrt{\sum_{i=0}^{q-1}\left(\frac{h}{2}f''(\xi_i)\right)^2}\leq\frac{h}{2}\cdot q\cdot \mathcal{H},
\end{align}
où $\mathcal{H}=\max_i{f''(\xi_i)}$. On peut également réécrire $q$ comme étant $q=(x_{q-1}-x_0)/h$. On a finalement
$$
E_\mathrm{tot}\leq \left(\frac{x_{q-1}-x_0}{2}\right)\cdot\mathcal{H}.
$$
De façon similaire, on obtient pour l'erreur moyenne
$$
E_\mathrm{moy}\leq \frac{h}{2}\mathcal{H}.
$$
On voit donc que l'erreur totale sera constante avec l'augmentation de $q$ (ou la diminution de $h$). En revanche l'erreur moyenne diminue linéairement avec $h$. La preuve en image.

![L'erreur moyenne et totale sur le calcul de la dérivée de $f(x)=\sin(x)$ ($f'(x)=\cos(x)$) en fonction de $h$.](figs/err_deriv_avant.svg){#fig:err_deriv_avant width=70%}

Sur la @fig:err_deriv_avant, on peut voir l'erreur totale, ainsi que l'erreur moyenne. Comme attendu l'erreur totale vaut environ un et est constante
en fonction de $h$. On voit également que l'erreur moyenne augmente linéairement avec $h$.

La @fig:err_deriv_avant est obtenue à l'aide du listing suivant

```python
import numpy as np
import matplotlib.pyplot as plt
import math

def f(x):
    return np.sin(x)

def df(x):
    return np.cos(x)

def app(x, h):
    return (f(x+h)-f(x)) / h

def comp_error_sqr(xmin, xmax, h):
    x = np.arange(xmin, xmax, h)
    approx = app(x,h)
    exact = df(x)

    error = (approx - exact)**2
    return x, error

def comp_error_moy(error):
    q = error.size
    return np.sum(np.sqrt(error) / q)

def comp_error_tot(error):
    return np.sum(np.sqrt(error))



h = 0.2
xmin = 0.0
xmax = math.pi

x, error = comp_error_sqr(xmin, xmax, h)

plt.figure(1)

plt.subplot(121)
plt.title('Fonction')
plt.plot(x, df(x), 'b-', label="$f'(x)=\cos(x)$")
plt.plot(x, app(x, h), 'rx-', label="approx.")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.plot(x, error, 'b-')
plt.ylabel('Erreur')
plt.xlabel('$x$')
plt.grid(True)
plt.show()

h = [0.025, 0.05, 0.1, 0.2, 0.4]
xmin = 0.0
xmax = math.pi

e_tot = np.array([])
e_moy = np.array([])

for dx in h:
    x, error = comp_error_sqr(xmin, xmax, dx)
    e_tot = np.append(e_tot, comp_error_tot(error))
    e_moy = np.append(e_moy, comp_error_moy(error))


plt.figure(2)

plt.plot(h, e_tot, 'b-', label="$E_\mathrm{tot}$")
plt.plot(h, e_moy, 'rx-', label="$E_\mathrm{moy}$")
plt.legend(loc=0)
plt.ylabel('Erreur')
plt.xlabel('$h$')
plt.grid(True)
plt.show()
```

### Méthode de différences finies centrales d'ordre 2

De façon assez similaire, à ce que nous avons fait pour les différences finies d'ordre 1 en avant (et en arrière), nous
pouvons évaluer l'erreur pour les différences d'ordre 2. A l'aide de l'@eq:err_tot et de l'@eq:err_moy, on peut calculer les erreurs moyennes et
totales. Pour ce faire nous devons étudier la fonction suivante
$$
E_\mathrm{tot}=\sqrt{\sum_{i=0}^{q-1}\left(f'(x_i)-\frac{f(x_i+h/2)-f(x_i-h/2)}{h}\right)^2}
$$.
Pour ce faire nous commençons par écrire la série de Taylor de $f(x_i+h/2)$ et $f(x_i-h/2)$
\begin{align}
f(x_i+h/2)&=f(x_i)+\frac{h}{2}f'(x_i)+\frac{h^2}{8}f''(x_i)+\frac{h^3}{48}f'''(\xi),\\
f(x_i-h/2)&=f(x_i)-\frac{h}{2}f'(x_i)+\frac{h^2}{8}f''(x_i)-\frac{h^3}{48}f'''(\eta).
\end{align}
En substituant ces deux relations dans la formule ci-dessus, il vient
\begin{align}
E_\mathrm{tot}&=\sqrt{\sum_{i=0}^{q-1}\left(f'(x_i)-\frac{hf'(x_0)+\frac{h^3}{48}\left(f'''(\xi)+f'''(\eta)\right)}{h}\right)^2},\nonumber\\
&=\frac{h^2}{48}\sqrt{\sum_{i=0}^{q-1}\left(\left(f'''(\xi)+f'''(\eta)\right)\right)^2}\leq \mathcal{H}\cdot h,
\end{align}
oû $\mathcal{H}$ est la valeur maximale de $f'''(\xi)+f'''(\eta)$. De même il vient que
$$
E_\mathrm{moy}=\leq \mathcal{H}\cdot h^2.
$$
On a donc que la formule des différences finies centrées est d'ordre 2, notée $\mathcal{O}(h^2)$. L'erreur moyenne décroissant proportionnellement à $h^2$: lorsqu'on divise $h$ par 2, on divise l'erreur par 4.

![L'erreur moyenne et totale sur le calcul de la dérivée de $f(x)=\sin(x)$ ($f'(x)=\cos(x)$) en fonction de $h$ pour les différences finies en avant
et celle centrées.](figs/err_deriv_avant_centree.svg){#fig:err_deriv_avant_centree width=70%}

On peut voir ce comportement sur la @fig:err_deriv_avant_centree que l'erreur est beaucoup plus faible à résolution équivalente pour
les différences finies centrées par rapport aux différences finies en avant. De plus la pente de l'erreur est plus faible
pour les différences finies centrées. Elle est en fait de pente $2$ en échelle log-log, contrairement aux différences finies centrées en avant
qui est de pente un.  

![L'erreur sur chaque point $x_i$ pour un $h$ fixé sur le calcul de la dérivée de $f(x)=\sin(x)$ ($f'(x)=\cos(x)$) en fonction de $h$ pour les différences finies en avant
et celle centrées (droite), ainsi que la valeur de $f'(x)$ pour chacune des deux méthodes. On constate que dans ce cas on ne peut pas distinguer $f'(x)$ de sa valeur approximée.](figs/err_cos_ac.svg){#fig:err_cos_ac width=70%}

On constate également sur la @fig:err_cos_ac que l'erreur sur chaque point est beaucoup plus faible poru les différences finies centrées, par rapport à celle en avant.
On doit en effet utiliser l'échelle logarithmique pour voir l'erreur des différences finies centrées, sinon on ne les verrait simplement pas sur le graphique (elle est de deux ordres
de grandeur plus faible).

Les @fig:err_deriv_avant_centree et @fig:err_cos_ac sont obtenues avec le listing suivant

```python
import numpy as np
import matplotlib.pyplot as plt
import math

def f(x):
    return np.sin(x)

def df(x):
    return np.cos(x)

def app(x, h):
    return (f(x+h)-f(x)) / h

def app_c(x, h):
    return (f(x+h/2)-f(x-h/2)) / h

def comp_error_sqr(xmin, xmax, h):
    x = np.arange(xmin, xmax, h)
    approx = app(x,h)
    exact = df(x)

    error = (approx - exact)**2
    return x, error

def comp_error_sqr_c(xmin, xmax, h):
    x = np.arange(xmin, xmax, h)
    approx = app_c(x,h)
    exact = df(x)

    error = (approx - exact)**2
    return x, error

def comp_error_moy(error):
    q = error.size
    return np.sum(np.sqrt(error) / q)

def comp_error_tot(error):
    return np.sum(np.sqrt(error))



h = 0.2
xmin = 0.0
xmax = math.pi

x, error = comp_error_sqr(xmin, xmax, h)
x, error_c = comp_error_sqr_c(xmin, xmax, h)

plt.figure(1)

plt.subplot(121)
plt.plot(x, df(x), 'b-', label="$f'(x)=\cos(x)$")
plt.plot(x, app(x, h), 'rx-', label="avant")
plt.plot(x, app_c(x, h), 'gx-', label="centr.")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.loglog(x, error, 'b-', label="Erreur en avant")
plt.loglog(x, error_c, 'g-', label="Erreur centrée")
plt.ylabel('Erreur')
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)
plt.show()

h = [0.025, 0.05, 0.1, 0.2, 0.4]
xmin = 0.0
xmax = math.pi

e_tot = np.array([])
e_tot_c = np.array([])
e_moy = np.array([])
e_moy_c = np.array([])

for dx in h:
    x, error = comp_error_sqr(xmin, xmax, dx)
    e_tot = np.append(e_tot, comp_error_tot(error))
    e_moy = np.append(e_moy, comp_error_moy(error))

    x, error = comp_error_sqr_c(xmin, xmax, dx)
    e_tot_c = np.append(e_tot_c, comp_error_tot(error))
    e_moy_c = np.append(e_moy_c, comp_error_moy(error))


plt.figure(2)

plt.loglog(h, e_tot, 'b-', label="$E_\mathrm{tot,avant}$")
plt.loglog(h, e_moy, 'rx-', label="$E_\mathrm{moy,avant.}$")
plt.loglog(h, e_tot_c, 'k-', label="$E_\mathrm{tot,centr.}$")
plt.loglog(h, e_moy_c, 'gx-', label="$E_\mathrm{moy,centr.}$")
plt.legend(loc=0)
plt.ylabel('Erreur')
plt.xlabel('$h$')
plt.grid(True)
plt.show()
```

---

Exercice (Ordre 3) +.#

Saurez-vous écrire la formule d'ordre $\mathcal{O}(h^3)$ pour calculer la dérivée d'une fonction?

---

## Que se passe-t-il quand $h\rightarrow 0$?{#sec:arroni_dervi}

Lorsque nous utilisons les formules précédentes sur un ordinateur, nous voudrions utiliser un $h$ aussi petit que possible afin d'être aussi précis que possible.
Néanmoins, plus $h$ devient petit, plus nous risquons des erreurs numériques. En effet, sur un ordinateur la précision numérique
à disposition est finie. En effet, les nombres à virgules flottantes se représentent en 32 ou 64 bits typiquement, soit une précision d'environ 7 chiffres significatifs et
16 respectivement. Ici, nous allons essayer de déterminer quelle est la valeur minimale de $h$ que nous pouvons utiliser en fonction du nombre de chiffres significatifs
que nous avons à disposition sur un ordinateur.

En d'autres termes soit un nombre réel $a$, et sa représentation sur un ordinateur, $\bar{a}$, la différence entre les deux est la précision numérique
de la machine et peut s'écrire comme
$$
|h-\bar{h}|\sim |h|\cdot 10^{-N},
$$
où $N$ est le nombre de chiffres significatifs de la machine. On a donc que $f(x_0+h)$ et $f(x_0)$ sont connus à
une précision
\begin{align}
f(x_0+h)-\overline{f(x_0+h)}\sim |f(x_0+h)|\cdot 10^{-N}\sim |f(x_0)|\cdot 10^{-N},\\
f(x_0)-\overline{f(x_0)}\sim |f(x_0)|\cdot 10^{-N}.
\end{align}
En utilisant cette information sur la formule de la dérivée en avant par exemple, on a
$$
f'(x_0)\cong \frac{f(x_0+h)-f(x_0)}{h}\sim \frac{2\cdot |f(x_0)|\cdot 10^{-N}}{h}.
$$
Si $h\gg  2\cdot |f(x_0)|\cdot 10^{-N}$ la précision numérique n'aura pas d'impact sur l'évaluation de la dérivée. En revanche, si
$h\sim 10^{-N}$, l'erreur d'arrondi aura un effet dramatique sur le résultat.

## Dérivées d'ordre deux

Nous souhaitons à présent calculer numériquement la deuxième dérivée de la fonction $f(x)$ en $x_0$, soit $f''(x_0)$.
Pour ce faire, il faut se rappeler que la deuxième dérivée d'une fonction, n'est autre que la dérivée de sa dérivée.
On peut donc utiliser la formule de différence finie centrée (par exemple) sur $f'(x)$[^1]
$$
f''(x_0)=\frac{f'(x_0+h/2)-f'(x_0-h/2)}{h}.
$${#eq:deriv_deux}
On peut à présent utiliser les formules de différences finies centrées sur $f'(x_0+h/2)$ et $f'(x_0-h/2)$
\begin{align}
f'(x_0+h/2)&=\frac{f(x_0+h/2+h/2)-f(x_0+h/2-h/2)}{h}=\frac{f(x_0+h)-f(x_0)}{h},\\
f'(x_0-h/2)&=\frac{f(x_0-h/2+h/2)-f(x_0-h/2-h/2)}{h}=\frac{f(x_0)-f(x_0-h)}{h}.
\end{align}
On peut constater que la première formule revient à calculer $f'(x_0)$ avec la formule de différences finies en avant,
alors que la seconde est celle en arrière. On peut donc réécrire l'@eq:deriv_deux comme
$$
f''(x_0)=\frac{f(x_0+h)-2f(x_0)+f(x_0-h)}{h^2}.
$$
A présent on peut évaluer l'erreur de cette approximation. En utilisant la formule de Taylor sur $f(x_0+h)$ et $f(x_0-h)$
\begin{align}
f(x_0+h)&=f(x_0)+h\cdot f'(x_0)+\frac{h^2}{2}\cdot f''(x_0)+\frac{h^3}{6}f'''(x_0)+\frac{h^4}{24}f^{(4)}(\xi),\\
f(x_0-h)&=f(x_0)-h\cdot f'(x_0)+\frac{h^2}{2}\cdot f''(x_0)+\frac{h^3}{6}f'''(x_0)+\frac{h^4}{24}f^{(4)}(\eta),
\end{align}
on peut écrire l'erreur
$$
\left|f''(x_0)-\frac{f(x_0+h)-2f(x_0)+f(x_0-h)}{h^2}\right|=\frac{h^2}{24}(f^{(4)}(\xi)+f^{(4)}(\eta))\leq \mathcal{H}\cdot h^2,
$$
oû $\mathcal{H}$ est la valeur maximale de $f'''(\xi)+f'''(\eta)$.

---

Exercice (En avant, puis en arrière) +.#

Déterminer la formule pour $f''(x_0)$ en utilisant la formule de la dérivée en avant à la place de celle centrée pour l'@eq:deriv_deux.

---

## Résumé

Dans ce chapitre, nous avons vu comment approximer le calcul de dérivées numériquement et prédire l'ordre de l'erreur de ces approximations.
Nous allons utiliser ces méthodes dans la section suivante où nous allons les utiliser pour interpoler des fonctions.
Nous avons également discuté l'effet des erreurs d'arrondi sur les dérivées.

# L'interpolation

Dans ce chapitre, nous allons étudier comment nous pouvons essayer de reconstruire de l'information manquante
dans un signal ou dans une image. Cela est très utile si nous souhaitons rééchantillonner un signal,
ou dépixeliser une image (un peu comme dans la série "les experts"). On appelle **techniques d'interpolation**
les méthodes utilisées pour résoudre ce genre de problème.

Dans la @fig:interp_experts nous pouvons voir 3 versions d'une image. La première pixelisée (tout à gauche),
la seconde (au centre) rééchantillonnée de façon naïve, et la troisième (à droite)
avec une méthode plus avancée.

![Différents rééchantillonages d'une image pixelisée (plus proche voisin, bilinéaire et bicubique de gauche à droite). Source: Wikipedia
<https://bit.ly/2wTidqj>, <https://bit.ly/2wTib1F>, et <https://bit.ly/2wPyjAA>.](figs/Interpolation-all.svg){#fig:interp_experts width=70%}

## Définitions diverses

Avant de faire à proprement parler des interpolations, nous allons d'abord poser le problème de façon
mathématique et aussi simple que possible.
Nous définissons un ensemble de $n$ points ($n\in\natural$) $\{y_i\}_{i=0}^{n-1}\in\real$,
qui correspondent à $n$ positions $\{x_i\}_{i=0}^{n-1}\in\real$ (voir @fig:points).
Pour simplifier, nous allons supposer qu'ils sont dans un ordre croissant, soit
\begin{equation}
x_0 < x_1< \cdots < x_{n-1}.
\end{equation}

![Six points $y_0$, ..., $y_5$ aux positions $x_0$, ..., $x_5$.](figs/points.svg){#fig:points width=100%}

Le but à présent est de déterminer une fonction $f$ définie
dans tout l'intervalle $[x_0,x_{n-1}]$ et passant par tous les points $y_i$ (voir @fig:points_int).

![La fonction d'interpolation $f(x)$ passant par les six paires de points $(x_0,y_0)$, ..., $(x_5,y_5)$.](figs/points_int.svg){#fig:points_int width=100%}

La fonction $f(x)$ n'est pas unique et il existe différents moyens de déterminer ces fonctions qui peuvent être plus ou moins complexes.
Dans les sections qui suivent, nous allons en voir quelques unes qui sont toutes basées sur les interpolations polynomiales.  C'est-à-dire,
sur une fonction $f(x)$ interpolante qui sera de la forme
\begin{equation}
p_m(x)=a_0+a_1\cdot x+a_2\cdot x^2+\dots+a_m\cdot x^m,
\end{equation}
où $p_m(x_i)=y_i$ ($i=0,...,n-1$), avec $m\in\natural$ et où tout le jeu consiste à trouver les valeurs des $\{a_i\}_{i=0}^m$.

## Idée naive

L'idée la plus simple que nous pouvons avoir pour déterminer un polynôme interpolant est
d'écrire le système d'équations suivant
\begin{align}
p_m(x_0)=y_0&=a_0+a_1x_0+\dots+a_mx_0^m,\\
p_m(x_1)=y_1&=a_0+a_1x_1+\dots+a_mx_1^m,\\
\cdots\nonumber\\
p_m(x_{n-1})=y_n&=a_0+a_1x_{n-1}+\dots+a_mx_{n-1}^m.
\end{align}
On se retrouve donc avec un système de $n$-équations à $m+1$ inconnues.

---

Question +.#

Quel est l'ordre du polynôme qu'on peut faire passer par tous les $n$ points $x_i,y_i$ (si tous les $x_i$ et $y_i$ sont différents les uns des autres)?

---

De façon générale pour que le système d'équations ci-dessus aient une solution unique, il faut que le nombre d'équations soit égale au nombre d'inconnues,
et donc le nombre de points doit être égal au degré du polynôme d'interpolation ($m=n-1$). Il existe néanmoins des cas particuliers où le degré du polynôme d'interpolation
peut être plus petit que le nombre de points.

---

Question +.#

Pouvez-vous trouver de tels exemples?

---

Dans le cas général, nous essayons de résoudre le système de $n$ équations à $n$ inconnues qui peut se mettre sous forme matricielle
\begin{equation}
\mat{X}\cdot \vec a=\vec y,
\end{equation}
où $\vec a$ est l'inconnue (nous recherchons les coefficients du polynôme) et
où $\mat{X}$, $\vec y$ et $\vec x$ sont donnés par
\begin{equation}
\mat{X} = \begin{pmatrix}
                         1 & x_{0} & x_{0}^2  & \cdots & x_0^{n-1} \\
                         1 & x_{1} & x_{1}^2  & \cdots & x_1^{n-1} \\
                          \vdots & \vdots  & \ddots & \vdots \\
                          1 & x_{n-1} & x_{n-1}^2  & \cdots & x_{n-1}^{n-1}
\end{pmatrix},
\vec y=
\begin{pmatrix}
       y_0\\
       y_1\\
      \vdots\\
      y_{n-1}
\end{pmatrix},
\vec a=
\begin{pmatrix}
       a_0\\
       a_1\\
      \vdots\\
      a_{n-1}
\end{pmatrix}.
\end{equation}
Afin de déterminer les coefficients $a_i$, nous devons résoudre ce système d'équations linéaires, qui possède une solution unique
seulement si $\det(\mat{X})\neq 0$. Nous n'allons pas résoudre ce système de façon générale, car cela serait une solution beaucoup trop
couteuse.

---

Question +.#

Quel est le degré du polynôme d'interpolation qui passe par $2$ points?

---

---

Exemple (La droite) +.#

Voyons à présent comment déterminer le polynôme passant par deux points, $(x_0,y_0)=(0,0)$ et $(x_1,y_1)=(1,1)$.
Avec la méthode naïve, nous pouvons écrire le système d'équations suivant
\begin{align}
a_0+a_1\cdot 0=0,\\
a_0+a_1\cdot 1=1.
\end{align}
La solution à ce système est $a_0=0$, et $a_1=1$. Le polynôme est donc donné par
\begin{equation}
p_1(x)=x.
\end{equation}

---

---

Exercice (La parabole) +.#

Déterminer le polyôme d'interpolation passant par les points
$(x_0,y_0)=(0,1)$, $(x_1,y_1)=(1,0)$, et $(x_2,y_2)=(-1,0)$.

---

La méthode naïve bien que très simple à écrire est beaucoup trop compliquée à résoudre. En effet, cela implique en général de résoudre
un système d'équations linéaire où $\mat{X}$ est une matrice dont tous les coefficients sont non-nuls. On doit donc employer des méthodes
comme la décomposition $LU$ pour y parvenir. Grâce à différentes autres techniques que nous pouvons voir dans ce qui suit,
nous pouvons grandement nous faciliter la tâche en écrivant le système d'équations sous une forme beaucoup plus
agréable à résoudre.

En fait si on écrit le polynôme $p_m(x)$ sous la forme[^2]
\begin{equation}
p_{m}(x)=b_0+b_1\cdot (x-x_0)+b_2\cdot (x-x_0)(x-x_1)+\dots+b_{m}\cdot (x-x_0)(x-x_1)\cdots(x-x_{m-1}),
\end{equation}
le système d'équations $p_{n-1}(x_i)=y_i$, pour $i=0,...,n-1$, devient
\begin{align}
p_{n-1}(x_0)&=b_0=y_0,\\
p_{n-1}(x_1)&=b_0+b_1(x_1-x_0)=y_1,\\
&\vdots\nonumber\\
p_{n-1}(x_{n-1})&=b_0+b_1(x_{n-1}-x_0)+b_2(x_{n-1}-x_0)(x_{n-1}-x_1)(x_{n-1}-x_2)\nonumber\\
&+\cdots b_{n-1}(x_{n-1}-x_0)(x_{n-1}-x_1)\cdots(x_{n-1}-x_{n-2}) =y_{n-1}.
\end{align}
On a donc un système d'équations à résoudre de la forme
$$
\mat{X}'\cdot \vec b=\vec y,
$$
où $\mat{X}'$ est une matrice triangulaire inférieure (seule la partie sous la diagonale est non-nulle)
\begin{equation}
\mat{X}'' =
\begin{pmatrix}
    1 & 0 & 0  & \cdots & 0 \\
    1 & x_{1}-x_0 & 0  & \cdots & 0 \\
    \vdots & \vdots  & \ddots & \vdots & 0\\
    1 & x_{n-1}-x_0 & (x_{n-1}-x_0)(x_{n-1}-x_1)  & \cdots & (x_{n-1}-x_0)(x_{n-1}-x_1)\cdots(x_{n-1}-x_{n-2})
\end{pmatrix},
\end{equation}
Le système d'équation est donc beaucoup plus simple à résoudre.

## Les différences divisées

Afin de construire un polynôme de façon un peu plus intuitive que ce que nous venons de faire, nous allons commencer par simplifier le problème. Considérons le cas très simple, où nous n'avons que deux points, $(x_0,y_0)$ et $(x_1,y_1)$. Nous pouvons faire passer un polynôme de degré un (une droite donc) par ces points, et sont équation sera
$$
p_1(x)=y_0+(x-x_0)\frac{y_1 - y_0}{x_1-x_0}.
$$
On constate aisément que le terme de droite dans cette équation n'est autre que l'approximation de la dérivée (d'ordre 1 en avant) d'une fonction quelconque, $f$, passant par les points $(x_0,y_0)$ et $(x_1,y_1)$
$$
f'(x_0)\cong \frac{y_1 - y_0}{x_1-x_0}.
$$
On a donc que $p_1(x)$ peut se mettre sous la forme
$$
p_1(x)=y_0+(x-x_0)f'(x_0).
$$
Ce polynôme ressemble terriblement au développement de Taylor de $f(x)$ proche de $x_0$.

Si à présent, nous avons $3$ points à disposition, soit
$(x_0,y_0)$, $(x_1,y_1)$, et $(x_2,y_2)$, nous pouvons faire passer un polynôme de degré deux par ces points. Comme nous connaissons déjà le polynôme de degré un passant par $(x_0,y_0)$ et $(x_1,y_1)$, nous
allons écrire $p_2(x)$ comme $p_1(x)$ auquel nous allons ajouter une correction
$$
p_2(x)=p_1(x)+b_2(x-x_0)(x-x_1),
$$
où $b_2$ est un coefficient à déterminer. La forme du terme de droite de l'équation ci-dessus impose que $p_2(x_0)=p_1(x_0)$ et $p_2(x_1)=p_1(x_1)$. On a donc trivialement que
\begin{align}
&p_2(x_0)=y_0,\\
&p_2(x_1)=y_1.
\end{align}

---

Exercice (La parabole, bis) +.#

Déterminer $a$.

---

Afin de calculer $a$, il nous reste donc l'équation $p_2(x_2)=y_2$ pour déterminer
$a$
$$
p_2(x_2)=y_0+(x_2-x_0)\frac{y_1-y_0}{x_1-x_0}+(x_2-x_1)(x_2-x_0)b_2=y_2.
$$
En soustrayant $p_2(x_1)$ et $p_2(x_2)$, puis en divisant le résultat par
$(x_2-x_0)(x_2-x_1)$, il vient
$$
b_2=\frac{1}{x_2-x_0}\left(\frac{y_2-y_1}{x_2-x_1}-\frac{y_1-y_0}{x_1-x_0}\right).
$$
On constate que $b_2$ est en fait une approximation de la deuxième dérivée d'une fonction passant $f$ par $(x_0,y_0)$, $(x_1,y_1)$, et $(x_2,y_2)$. En effet
\begin{align}
f'(x_2)&\cong\frac{y_2-y_1}{x_2-x_1},\\
f'(x_0)&\cong\frac{y_1-y_0}{x_1-x_0},
\end{align}
où à la première ligne on évalue la dérivée "en arrière" et vers l'avant sur la deuxième. Il nous reste à voir qu'avec ces définitions
$b_2$ est donnée par
$$
b_2=\cong\frac{f'(x_2)-f'(x_0)}{x_2-x_0}\cong f''(x_0),
$$
qui n'est autre que la formule de la dérivation vers l'avant d'ordre un pour $f'(x)$. Finalement, on a pour $p_2(x)$
$$
p_2(x)=y_0+(x-x_0)\frac{y_1 - y_0}{x_1-x_0}+(x-x_0)(x-x_1)\left(\frac{\frac{y_2-y_1}{x_2-x_1}-\frac{y_1-y_0}{x_1-x_0}}{x_2-x_0}\right).
$$
Cette façon de construire les polynômes d'interpolation se généralise.
On peut définir les différences divisées comme
\begin{align}
y[x_i]&=y_i,\\
\delta y[x_i,x_j]&=\frac{y[x_j]-y[x_i]}{x_j-x_i},\\
\delta^2 y[x_i,x_j,x_k]&=\frac{\delta y[x_j,x_k]-\delta y[x_i,x_j]}{x_k-x_i},\\
&\vdots\nonumber\\
\delta^n y[x_{i_0},x_{i_1},\dots,x_{i_n}]&=\frac{\delta^{n-1} y[x_{i_1},\dots,x_{i_{n}}]-\delta^{n-1} y[x_{i_0},\dots,x_{i_{n-1}}]}{x_{i_n}-x_{i_0}}.
\end{align}
Avec cette définition, on peut définir le polynôme interpolant
passant par les points $(x_i,y_i)$, pour $i=0,..,n$ avec la **formule de Newton**, comme
\begin{align}
p_n(x)&=y[x_0]+(x-x_0)\delta y[x_0,x_1]+\dots\nonumber\\
&\quad+(x-x_0)(x-x_1)\cdots(x-x_{n-1})\delta y^n[x_0,\dots,x_n].
\end{align}

---

Exercice (Polynôme de degré 2) +.#

Soient les points $(x_i,y_i)$ donnés par $x_0=0$, $x_1=1$, $x_2=-1$, $y_0=0$, $y_1=1$, $y_2=1$. Écrire les différences divisées et le polynôme d'interpolation passant par ces trois points.

---

### Erreur de l'interpolation

Nous avons vu dans la section précédente comment écrire un polynôme d'interpolation passant par $n+1$ points, $(x_i,y_i)$, avec $x_i\in[a,b]$, pour $a,b\in\real$. Si nous supposons que
$$
y_i=f(x_i),
$$
où $f$ est une fonction
$$
f:[a,b]\rightarrow\real,
$$
dérivable au moins $n$ fois et que
les points $(x_i,y_i)$ ne sont en fait qu'un échantillonnage de $f(x)$.
Nous pouvons nous poser la question de ce que vaudra la différence entre $p_n(x)$ et $f(x)$.
En particulier, est-ce que si $n\rightarrow\infty$ on a bien que $p_n(x)\rightarrow f(x)$?

La réponse à cette question n'est pas aisée, pour s'en rendre compte prenons deux fonctions différentes
\begin{align}
f(x)&=\sin(x),\ x\in[-5,5],\\
g(x)&=\frac{1}{1+x^2},\ x\in[-5,5].
\end{align}
Pour la fonction $f(x)$, on voit sur la @fig:sin_interp_n que lorsqu'on fait passer le nombre de point d'interpolation de $n=5$ à $n=20$,
le polynôme d'interpolation s'approche de plus en plus de $\sin(x)$ jusqu'à ce qu'on ne puisse plus distinguer de différences.
Bonne nouvelle donc!

![Le polynôme d'interpolation (en rouge) passant par les points noirs, qui échantillonnent la fonction $f(x)=\sin(x)$ pour $n=5$ points (à gauche)
et $n=20$ à droite.](figs/sin_interp_n.svg){#fig:sin_interp_n width=100%}

En revanche pour la fonction $g(x)$, on voit sur la @fig:one_over_one_x2 qu'en augmentant $n=5$ à $n=20$, le polynôme d'interpolation
s'approche de plus en plus de $g(x)$ proche du centre du domaine ($x\cong0$), mais que sur les bords,
l'interpolation devient de plus en plus mauvaise. Ceci est donc une mauvaise nouvelle!

![Le polynôme d'interpolation (en rouge) passant par les points noirs, qui échantillonnent la fonction $g(x)=\frac{1}{1+x^2}$ pour $n=5$ points (à gauche)
et $n=20$ à droite.](figs/one_over_one_x2.svg){#fig:one_over_one_x2 width=100%}

En fait comme vous pouvez le constater ici, nous ne sommes pas assurés d'avoir une convergence du polynôme d'interpolation vers la fonction
$f(x)$ en général. Il se trouve que la différence entre le polynôme d'interpolation, $p_n(x)$ et $f(x)$ est donnée par[^3]
$$
f(x)-p_n(x)=(x-x_0)\cdot (x-x_1)\cdot \dots \cdot (x-x_n)\cdot \frac{f^{(n+1)}(\xi)}{(n+1)!},
$$
où $\xi\in[a,b]$. On voit donc que si $f^{(n+1)}(\xi)\geq (n+1)!$ on va avoir une croissance de la différence entre $p_n(x)$ et $(x)$, sinon tout va bien se passer. Regardons ce qui se passe pour le $f(x)=\sin(x)$ d'abord. Ses dérivées sont données par
$$
f^{(2n)}=\pm\sin(x)\in[-1,1],\quad f^{(2n+1)}=\pm\cos(x)\in[-1,1].
$$
On a donc que la différence va tendre vers zéro pour $n\rightarrow\infty$.
En revanche pour $g(x)=1/(1+x^2)$, les dérivées se comportent beaucoup moins bien. Par exemple,
$$
g^{(10)}=\frac{3628800\cdot(11 x^{10}-165 x^8+462 x^6-330 x^4+55 x^2-1))}{(x^2+1)^{11}}.
$$
On voit donc que $g^{(10)}= 11!\cdot \frac{(x^{10}-15 x^8+42 x^6-30 x^4+5 x^2-\frac{1}{11})}{(x^2+1)^{11}}$. Cette relation se généralise et
on a que
$$
g^{(n)}\sim (n+1)!.
$$
Cette relation implique que la différence entre $g(x)$ et $p_n(x)$ va croître avec $n$
\begin{align}
g(x)-p_n(x)&\sim(x-x_0)\cdot (x-x_1)\cdot \dots \cdot (x-x_n)\cdot \frac{(n+2)!}{(n+1)!}\nonumber\\
&\sim(x-x_0)\cdot (x-x_1)\cdot \dots \cdot (x-x_n)\cdot(n+2).
\end{align}
Ce phénomène de la croissance de l'erreur avec l'augmentation de l'ordre de l'interpolation est appelé *phénomène de Runge*.

En fait ce comportement de l'erreur est dû au fait que les points que nous avons choisi pour l'interpolation sont **équidistants**.
On peut améliorer grandement le comportement de l'interpolation en rajoutant des points proches des bords du domaine d'interpolation.
Une solution est d'utiliser l'interpolation de Chebyshev que nous ne verrons pas dans le cadre de ce cours.

### Erreur d'arrondi

Une autre source d'erreur qui peut apparaître lorsque nous faisons des interpolations, est l'erreur d'arrondi (comme nous l'avions discuté
dans le chapitre précédent pour les dérivées, voir @sec:arroni_dervi).

Sans entrer dans les détails de l'évaluation de l'effet de l'erreur d'arrondi, on peut simplement le constater sur l'exemple de $f(x)=\sin(x)$. Nous savons de ce que nous avons discuté dans la section précédente, que l'interpolation de $f(x)=\sin(x)$ va bien se passer, car ses dérivées sont bornées. Sur la @fig:sin_arrondi, on constate que lorsqu'on augmente le nombre de points d'interpolation, on a soudain une augmentation dramatique de l'erreur.

![Le polynôme d'interpolation (en rouge) passant par les points noirs, qui échantillonnent la fonction $f(x)=\sin(x)$ pour $n=50$ (gauche), $n=75$ (milieu), et $n=100$ (droite).](figs/sin_arrondi.svg){#fig:sin_arrondi width=100%}

## Interpolation par intervalles

Outre utiliser d'autres polynômes pour effectuer l'interpolation, il existe d'autres méthodes
pour s'affranchir du phénomène d'oscillations de Runge. Pour ce faire, il suffit d'utiliser des fonctions polynomiales plus simples (de degrés $n$ plus faibles) qui seront utilisées sur des sous-ensembles des points $(x_i, y_i)$.

### Interpolation par intervalles de degré un

On considère toujours un ensemble $N$ de points équidistants
entre deux bornes $a,b\in\real$ et $x_i=a+h\cdot i$, avec $h=\frac{b-a}{N}$. Sur chacun des points $x_i$, nous avons une fonction $f(x_i)$ passant par les points $y_i$
$$
f(x_i)=y_i.
$$
On veut construire une fonction $f_h(x)$, continue, qui passe par les points $(x_i,y_i)$ et qui est une droite (un polynôme de degré 1, $p_1(x)$) sur chaque intervalle $[x_i,x_{i+1}]$. On va donc avoir que
$$
f_h(x_i)=y_i,\mbox{ et } f_h(x)\in p_1(x),\ x\in[x_i,x_{i+1}],\ i=1,..,N-1.
$$
Un exemple d'une telle fonction peut se voir sur la @fig:sin_deg1.

![L'interpolation par intervalle sur la fonction $f(x)=\sin(x)$ (en noir) sur $n=10$ points et $f_h(x)$ (en rouge) qui est un polynôme de degré un par morceaux sur $[x_i,x_{i+1}]$.](figs/sin_deg1.svg){#fig:sin_deg1 width=50%}

Ici, comme on peut le voir sur le listing ci-dessous qui a été utilisé pour générer la @fig:sin_deg1, l'interpolation par intervalle de degré un est l'interpolation par défaut utilisée par la librairie python, `matplotlib` lors de l'affichage de graphiques.

```{.python}
import numpy as np
import matplotlib.pyplot as plt
import math

def echant(x0, x1, n, f):
    dx = (x1-x0)/n
    x = np.arange(x0,x1+dx/10,dx)
    # return (x,np.sin(x))
    return (x,f(x))

n = 5
x0 = -5
x1 = 5
(x,y) = echant(x0, x1, n, np.sin)

plt.plot(x,y,'ro-')


axes = plt.gca()
axes.set_ylim([-1.1,1.1])

n *= 100
(x,y) = echant(x0, x1, n, np.sin)
plt.plot(x,y,'k-')
plt.xlabel("$x$")
plt.ylabel("$\sin(x)$")
plt.grid(True)
plt.show()
```

---

Remarque +.#

De façon générale cette fonction sera continue, mais pas dérivable (comme on peut le constater sur la @fig:sin_deg1) .

---

Nous voulons à présent savoir si notre fonction $f_h(x)\rightarrow f(x)$, pour $h\rightarrow 0$ (ce qui est équivalent à $N\rightarrow \infty$).

En fait un théorème que nous ne démontrerons pas, nous dit que

*Il existe un $C\in \real$, tel que pour tout $f:[a,b]\rightarrow\real$ deux fois continuement dérivable, et $\forall h>0$, l'erreur entre $f(x)$ et $f_h(x)$ est bornée par
$$
\max_{a\leq x\leq b}|f(x)-f_h(x)|\leq Ch^2\max_{a\leq x\leq b}|f''(x)|.
$$*

En fait cela revient à dire que si $f(x)$ est deux fois continuement dérivable, alors lorsque nous divisons $h$ par deux, alors l'erreur est diminuée par un facteur quatre au moins. De façon, similaire à ce que nous avons vu dans le chapitre précédent, on peut dire que l'interpolation par intervalle de degré un est d'ordre deux.

<!-- Cette évaluation de l'erreur est valable sur chaque point $x$. Si on veut étudier l'erreur totale dans un intervalle $[a,b]$, il est nécessaire de calculer "l'intégrale" (la somme) des erreurs sur tout l'intervalle.
\begin{align}
E&=\sum_{j=0}^{N-1} |f(x_j)-f_h(x_j)|\dd x\leq \sum_{j=0}^{N-1} \max_{a\leq x\leq b}|f(x)-f_h(x)|\nonumber\\
&\leq \sum_{j=0}^{N-1} Ch^2\max_{a\leq x\leq b}|f''(x)|
=N Ch^2\max_{a\leq x\leq b}|f''(x)|\nonumber\\
&=\frac{b-a}{h}Ch^2\max_{a\leq x\leq b}|f''(x)|\sim\mathcal{O}(h).
\end{align}
On constate donc que localement l'erreur est proportionnelle à $h^2$ (en chaque point) et globalement à $h$. -->

Nous savons maintenant que $f_h(x)$ tend vers $f(x)$ (tant que $f(x)$ nous fait la gentillesse d'être deux fois continuement dérivable). Voyons à présent comment on construit $f_h(x)$. Nous savons que pour $x\in[x_i,x_{i+1}]$, $f_h(x)$ passe par les points $(x_i,y_i)$ et $(x_{i+1}, y_{i+1})$, on peut donc écrire le système d'équations
\begin{align}
f_h(x_i)&=a_i\cdot x_i+b_i=y_i,\\
f_h(x_{i+1})&=a_i\cdot x_{i+1}+b_i=y_{i+1},
\end{align}
où $a_i$ et $b_i$ sont la pente et l'ordonnée à l'origine du segment de droite passant par $(x_i,y_i)$ et $(x_{i+1},y_{i+1})$.
Ce système de deux équations à deux inconnues se résout assez facilement en soustrayant la première de la deuxième équation, et on obtient directement $a_i$
$$
a_i=\frac{y_{i+1}-y_i}{x_{i+1}-x_i},
$$
et on obtient ensuite pour $b_i$
$$
b_i=-\frac{x_iy_{i+1}-x_{i+1}y_i}{x_{i+1}-x_i}.
$$
On a donc finalement que pour $i=0,..,N-1$, $f_h(x)$ est donné par
$$
f_h(x)=a_ix+b_i,\quad x\in[x_i,x_{i+1}].
$$

On peut même faire un exemple pour voir si cela est observable dans un cas particulier. Soit $f(x)=\sin(x)$, le listing suivant calcule l'erreur pour $N=10, ..., 100$ par pas de $10$.

```{.python}
import numpy as np
import matplotlib.pyplot as plt
import math

def echant(x0, x1, n, f):
    dx = (x1-x0)/n
    x = np.arange(x0,x1+dx/10,dx)
    return (x,f(x))

def interp_piece(x, x_p1, y, y_p1, dx):
    a = (y_p1 - y) / (x_p1 - x)
    b = - (x*y_p1 - x_p1 * y) / (x_p1-x)

    loc_x = np.arange(x, x_p1, dx)

    return (loc_x, loc_x * a + b)

def interp(x0, x1, n1, n2, f):
    dx = (x1 - x0) / n1
    all_x = np.array([])
    all_y = np.array([])
    for i in range(n1):
        x = x0 + i*dx
        x_p1 = x0 + (i+1)*dx

        loc_dx = (x_p1 - x) / n2
        (loc_x, loc_y) = interp_piece(x, x_p1, f(x), f(x_p1), loc_dx)

        if np.size(loc_x) > n1: # sometimes ranges goes too far
            loc_x = loc_x[0:n1]
            loc_y = loc_y[0:n1]


        all_x = np.append(all_x, loc_x)
        all_y = np.append(all_y, loc_y)

    return (all_x, all_y)

res = np.array([])
error = np.array([])
for n in range(10,100,10):
    x0 = -5
    x1 = 5
    (x,y) = echant(x0, x1, n, np.sin)

    (x,y) = interp(x0, x1, n, n, np.sin)

    n *= n
    (x_a,y_a) = echant(x0, x1, n, np.sin)

    error = np.append(error, np.sqrt(np.sum((y_a[0:n]-y)**2) / n))
    res = np.append(res, math.sqrt(n))

plt.loglog(res, error, 'ko-', label="Error")
plt.loglog(res, res**(-2), 'g-', label="$N^{-2}$")
plt.xlabel("$N$")
plt.legend(loc=0)
plt.grid(True)
plt.show()
```

On constate sur la @fig:error_deg1 que l'erreur est bien d'ordre deux: elle décroît à la même vitesse (elle a la même pente en échelle log-log) que $N^{-2}=\frac1/h^{-2}=h^2$.

![L'erreur en échelle log-log entre la fonction $\sin(x)$ et l'interpolation par morceaux de degrés en fonction de $N=1/h$ (en noir). A titre de comparaison, on a également tracé $N^{-2}$ en vert.](figs/error_deg1.svg){#fig:error_deg1 width=100%}

---

Exercice (Interpolation de degré deux par morceaux) +.#

Déterminer la forme de la fonction d'interpolation de degré deux par morceaux. A votre avis quel sera l'ordre de l'erreur?

---

### Spline

Les fonctions interpolantes par morceaux comme définies ci-dessus ont un désavantage: elles possèdent des
dérivées qui sont discontinues. Pour palier à ce problème, on a créé une autre catégorie
de polynômes d'interpolation: les *spline*. Le but est de trouver une fonction, $s:[a,b]\rightarrow \real$, qui passe par les points $(x_i,y_i)$
et qui est deux fois continuement différentiable (elle est dérivable partout deux fois et les dérivées sont continues)[^4].

Il se trouve qu'une telle fonction est une fonction cubique par morceaux sur chaque intervalle $[x_i,x_{i+1}]$.
Notons le polynôme défini sur $[x_i,x_{i+1}]$ comme $s_i(x)$. Nous voulons qu'il passe par les points $(x_i,y_i)$ et
$(x_{i+1},y_{i+1})$ et que sa dérivée soit continue. Notons $p_i$ la dérivée sur les points $x_i$ (la pente de la tangente). On a donc quatre conditions
sur l'intervalle $[x_i,x_{i+1}]$ que $s_i(x)$ doit satisfaire
\begin{align}
s_i(x_i)&=y_i,\\
s_i(x_{i+1})&=y_{i+1},\\
s_i'(x_i)&=p_i,\\
s_i'(x_{i+1})&=p_{i+1}.
\end{align}
On a donc également que $s_{i-1}'(x_i)=p_i$ et $s_{i+1}'(x_{i+1})=p_{i+1}$. Nous savons que $s_i(x)$ passe par $y_i$ et $y_{i+1}$
en $x_i$ et $x_{i+1}$, nous connaissons donc ses deux premiers termes grâce aux différences divisées.
On a $s_i(x)=y_i+\delta y[x_{i+1},x_i]+...$. Les termes suivants, sont de la forme suivante
\begin{align}
s_i(x)&=y_i+\delta y[x_{i+1},x_i](x-x_i)+c(x-x_i)(x-x_{i+1})\nonumber\\
&\quad\quad+d\left((x-x_i)^2(x-x_{i+1})+(x-x_i)(x-x_{i+1})^2\right),
\end{align}
où $c$ et $d$ sont deux coefficients à déterminer. Le deuxième terme (proportionnel à $(x-x_i)(x-x_{i+1})$) est la forme standard des différences divisées de Newton, et le troisième est le terme s'annulant en $x_i$ et $x_{i+1}$ mais étant symétrique
(on ne veut pas de terme du type $(x-x_i)^2(x-x_{i+1})$ car cela romprait la symétrie). Avec les conditions $s_i'(x_i)=p_i$ et $s'_i(x_{i+1})=p_{i+1}$, on peut trouver
les coefficients $c$ et $d$ en résolvant
\begin{align}
\delta y[x_{i+1},x_i]-c \left( x_{{i+1}}-x_{{i}} \right) +d \left( x_{{i+1}}-x_{{i}}
 \right) ^{2}=p_{{i}},\\
\delta y[x_{i+1},x_i]+c \left( x_{{i+1}}-x_{{i}} \right) +d \left( x_{{i+1}}-x_{{i}}
 \right) ^{2}=p_{{i+1}}.
\end{align}
On trouve aisément que
\begin{align}
c&=\frac{1}{2}\left(\frac{p_{i+1}-p_i}{x_{i+1}-x_i}\right),\\
d&=\frac{1}{2}\frac{1}{(x_{i+1}-x_i)^2}\left(p_{i+1}+p_i-2\delta y[x_{i+1},x_i]\right).
\end{align}
On constate donc que $c$ est proportionnel à la deuxième dérivée d'une fonction ayant pour dérivée $p_{i}$ et $p_{i+1}$ en $x_i$ et $x_{i+1}$ et que
$d$ est proportionnel à la deuxième dérivée d'une fonction ayant pour dérivée $p_{i}$, $p_{i+1}$, et $\delta y[x_{i+1},x_i]$ en $x_i$, $x_{i+1}$, et $x_{i+1/2}$.
On a que $s_i(x)$ est donné par
\begin{align}
s_i(x)&=y_i+\delta y[x_{i+1},x_i](x-x_i)+\frac{1}{2(x_{i+1}-x_i)}\left(p_{i+1}-p_i\right)(x-x_i)(x-x_{i+1})\nonumber\\
&\quad\quad+\frac{1}{2(x_{i+1}-x_i)^2}\left(p_{i+1}+p_i-2\delta y[x_{i+1},x_i]\right)\left((x-x_i)^2(x-x_{i+1})+(x-x_i)(x-x_{i+1})^2\right),
\end{align}

Nous avons à présent construit notre polynôme qui dépend de $x_i$, $x_{i+1}$, $y_i$, $y_{i+1}$, mais également de $p_i$ et $p_{i+1}$ que nous ne connaissons pas.
Pour déterminer ces deux grandeurs nous allons utiliser le fait que la deuxième dérivée de $s_i$ doit être continue, et donc
$$
s_i''(x_{i})=s_{i-1}''(x_{i}),\ i=1,..,N-1.
$$
Étant donné que $s_i''(x)$ est donné par
$$
s_i''(x)=\frac{p_{i+1}-p_{i}}{x_{i+1}-x_{i}}+3(2x-x_{i+1}-x_{i})\frac{(p_{i}+p_{i+1})(x_{i+1}-x_{i})+2(x_{i+1}-x_i)\delta y[x_{i+1},x_{i}]}{(x_{i+1}-x_{i})^3},
$$
on peut utiliser la condition ci-dessus pour écrire
$$
\frac{p_{i-1}}{x_i-x_{i-1}}+2p_i\left(\frac{1}{x_i-x_{i-1}}+\frac{1}{x_{i+1}-x_i}\right)+\frac{p_{i+1}}{x_{i+1}-x_i}=3\left(\frac{\delta y[x_{i+1},x_i]}{x_{i+1}-x_i}+\frac{\delta y[x_i,x_{i-1}]}{x_i-x_{i-1}}\right),
$$
En se souvenant que $x_{i+1}-x_i=x_i-x_{i-1}$ (les points sont équidistants) on peut même simplifier un peu plus
$$
p_{i-1}+4p_i+p_{i+1}=3(\delta y[x_{i+1},x_i]+\delta y[x_i,x_{i-1}]), \quad i=1,...,N-1.
$$
On peut donc écrire un système de $N-1$ équations à $N+1$ inconnues, $p_0$, ..., $p_N$. Il nous manque en effet
$p_0$ et $p_N$ qui doivent être fixés de façon ad-hoc (et on restera donc avec $N-1$ inconnues). Il y a trois cas principaux différents:

1. Le spline naturel: $s''(a)=s''(b)=0$.
2. Le spline scellé: $s'(a)=p_0$, $s'(b)=p_N$.
3. Le spline périodique: $s'(a)=s'(b)$, $s''(a)=s''(b)$.

Pour chacune de ces conditions cela va imposer une valeur pour $p_0$ et $p_N$ (dans le cas 2 elles sont imposées explicitement, dans les deux autres cas elles sont déduites).

# Optimisation

## Le voyageur de commerce

Le problème du voyageur du commerce est assez simple à formuler. Soit une liste de villes et les distances entre chaque paire connue. Quelle est
le parcours passant par toutes les villes une seule fois et arrivant dans la ville de départ minimisant la distance totale parcourue?

---

Exemple +.#

Soient les villes $A$, $B$, et $C$ comme sur la @fig:voyage_commerce. On peut comparer tous les chemins possibles

![Illustration des chemins entre les villes $A$, $B$, et $C$, ainsi que les distance les séparant.](figs/voyage_commerce.svg){#fig:voyage_commerce width=70%}

+------+------+------+------+------+------+
| ABC  | ACB  | BAC  | BCA  | CAB  | CBA  |
+------+------+------+------+------+------+
| 6 km | 6 km | 6 km | 6 km | 6 km | 6 km |
+------+------+------+------+------+------+

On constate d'abord que tous les trajets donnent le même résultat.
Essayons de comprendre pourquoi. La ville de départ ne change pas la longueur du chemin. On a donc que les chemins
"cycliques" ABC, BCA, CAB sont les mêmes chemins. De plus, comme les chemins peuvent être parcourus dans les deux sens,
on a également CBA, ACB et BAC sont les mêmes chemins. Il nous reste qu'un seul chemin possible.

---

---

Exercice +.#

![Illustration des chemins entre les villes $A$, $B$, $C$, et $D$ ainsi que les distance les séparant.](figs/voyage_commerce_4.svg){#fig:voyage_commerce_4 width=70%}

1. Combien de chemins possibles différents existe-t-il pour quatre villes $A$, $B$, $C$, et $D$?
2. Quel est le chemin le plus cours, si les distances entre les villes sont comme celles représentées sur la @fig:voyage_commerce_4?
3. Pouvez-vous généraliser le nomre de chemin à $N$ villes?

---

Le nombre total de chemin entre $n$ villes est de $n!$. Comme la ville de départ ne change pas le chemin à emprunter, on peut l'enlever du compte total,
il nous en reste donc $(n-1)!$. Finalement comme la direction dans laquelle nous parcourons le chemin ne change pas la distance,
nous pouvons encore diviser ce résultat par deux. Le nombre total de chemins existants est de $(n-1)!/2$. En supposant que le
calcul pour 5 villes prend $10\mu$s à calculer sur un ordinateur, on peut voir sur le tableau ci-dessous le temps que cela prendrait pour
10-25 villes. On constante le phénomène *d'explosion combinatoire* rendant le problème incalculable dans un temps raisonnable.

+--------------+-----------------+---------------------+
| Nb de villes | Nb possibilités | Temps (est.)        |
+:============:+:===============:+:===================:+
| 5            | 12              | $10\mu$s           |
+--------------+-----------------+---------------------+
| 10           | 181440          | 0.15s               |
+--------------+-----------------+---------------------+
| 15           | $4.4\cdot10^{10}$   | 10h                 |
+--------------+-----------------+---------------------+
| 20           | $6\cdot10^{16}$ | 1600 ans            |
+--------------+-----------------+---------------------+
| 25           | $3\cdot10^{23}$ | 8 milliard d'années |
+--------------+-----------------+---------------------+

La force brute est donc totalement hors de question pour résoudre ce type de problème. En fait il n'existe aucun algorithme **exact**
calculant la solution en temps polynomial. Les meilleurs algorithmes de ce types sont limités à 200'000 villes jusqu'à aujourd'hui. Néanmoins, il existe différentes heuristiques qui nous donnent une solution approchée (qui se trouve à 2-3% de la solution optimale avecune très grande probabilité)
et qui fonctionnent avec des millions de villes.

Le problème du voyageur du commerce est un problème **d'optimisation dicrète**, et en particulier **d'optimisation combinatoire**.
Ce genre de problème est en général beaucoup plus difficile à résoudre que les
problèmes d'optimisation continue bien que beaucoup plus simples à énoncer.
Une des difficulté est le cadre théorique (la formulation mathématique générale d'un problème d'optimisation discret) qui est
assez difficile à appréhender et qui peut être très différent entre les différents problèmes
d'optimisations.

Le voyageur du commerce peut s'écrire mathématiquement sous la formue suivante.
Supposons qu'il y ait $N$ villes qui sont numérotées de $1$ à $N$.
Soit $d_{ij}$ la distance entre la ville $i$ et la ville $j$ (on a que $d_{ij}d_{ji}$) pour $i,j\in 1,..,N$.
Soit $x_{ij}$ définiti comme
$$x_{ij}=
    \left\{\begin{array}{ll}
                $1,$&$\mbox{ si le chemin va de la ville } i { à la ville }j$\\
                $0,$&$\mbox{ sinon.}$
               \end{array}\right.$$

On cherche donc
\begin{align}
&\min \sum_{i=1}^N\sum_{j=1, j\neq i}^N x_{ij}d_{ij}, \mbox{ tel que},\\
&\sum_{i=1,i\neq j}^N x_{ij}=1,\ \forall i=1,..,N,\\
&\sum_{j=1,neq i}^N x_{ij}=1,\ \forall j=1,..,N.
\end{align}
La première contrainte nous dit qu'on ne peut arriver sur une ville $j$ que depuis une seule ville $i$ (impossible d'y arriver de deux façons différentes). La deuxième contrainte
nous dit que depuis chaque ville on ne peut partir qu'en direction d'une seule autre ville. Ces conditions (déjà assez complexes à formuler) ne suffisent pas pour définir le problème du voyageur de commerce. Il faut y ajouter deux contraintes supplémentaires
afin de garantir qu'il n'y a qu'un parcours qui couvre toutes les villes en y passant qu'une fois et que le parcours n'est pas une combinaison d'au moins deux parcours disjoints parcourant ensemble toutes les villes. Cette contrainte supplémentaire peut s'écrire (cette contrainte n'est pas exprimable de façon unique)
\begin{align}
&u_1=1,\\
&2\leq u_i\leq n,\ \forall i\neq 1,\\
&u_i-u_j+N\cdot x_{ij}\leq N-1,\quad i\neq 1,\ \j\neq 1,
\end{align}
où $u_i$ est une variable supplémentaire associée à chaque ville dont le seul but est
d'empêcher des cycles à l'intérieur du parcours (imposer qu'on passe qu'une seule fois par chaque ville).
Il est possible de s'en convaincre de la façon suivante. Cette contrainte supplémentaire nous impose que pour chaque
$x_{ij}=1$, on a que $u_j\geq u_i+1$. Par conséquent si on rentre dans une partie du parcours où on
rentre dans des sous-tours, les $u_i$ vont s'incrémenter à l'infini. Bien que les $u_i$ rajoutent des variables au problèmes (et pourraient le complexifier à première vue), on constate que les $u_i$ numérotent
les villes dans le parcours et peuvent aider à formuler un ordre ou la proximité d'une ville dans la suite.

## La régression linéaire

Lors d'une régression linéaire, le but est de trouver la droite, $y(x)=a\cdot x + b$, qui passe au mieux au travers d'un nuage de $N$ points $(x_i, y_i)$,
$i=1,...,N$ (voir @fig:reg).

![Nous recherchons l'équation de la droite, en rouge, $y(x)=ax+b$ passant au plus proche des points $(x_i,y_i)$ et bleu. Source: Wikipedia
<https://bit.ly/2SfiLzb>](figs/Linear_regression.svg){#fig:reg width=70%}

Pour déterminer l'équation de cette droite, nous devons donc trouver les coefficients $a$ et $b$ tels que la droite
passe au plus proche des points. Nous devons d'abord définir ce que signifie mathématiquement "passe au mieux par au travaers du nuage de points".
Une façon de mesurer la "qualité" d'une droite est de mesurer la somme des distances au carré entre les points $(x_i,y_i)$ et
la droite $y(x)=a\cdot x + b$ pour des valeurs de $a$ et $b$ données, soit
$$
E(a,b)=\sum_{i=1}^N (y(x_i)-y_i)^2.
$$
Nous cherchons par conséquent à minimiser $E(a,b)$ sous la contrainte que $y(x)$ est une droite. Pour simplifier encore plus le problème mathématique,
nous pouvons rajouter comme contrainte que la droite $y(x)$ passe par le point $(0,0)$, on a donc que $y(x)=a\cdot x$ (l'ordonnée à l'origine est nulle, $b=0$) et que
$$
E(a)=\sum_{i=1}^N (y(x_i)-y_i)^2,
$$
est indépendant de $b$. En résumé nous cherchons à résoudre le problème mathématique
\begin{align}
&\min_{a\in\real} E(a) = \min_{a \in\real} \sum_{i=1}^N (y(x_i)-y_i)^2,\\
&\mbox{où }y(x)=a\cdot x, \quad \mbox{(contrainte)}.
\end{align}
On peut réécrire la fonction $E(a)$ comme
\begin{align}
E(a)&=\sum_{i=1}^N \left(y^2(x_i)-2\cdot y_i\cdot y(x_i)+y_i^2\right)=\sum_{i=1}^N \left(a^2\cdot x_i^2-2\cdot a\cdot x_i\cdot y_i+y_i^2\right),\nonumber\\
    &=a^2\sum_{i=1}^Nx_i^2 + 2a\sum_{i=1}^Nx_iy_i+\sum_{i=1}^Ny_i^2.
\end{align}
Les $x_i$ et $y_i$ étant connus, nous cherchons $a$, tel que $E(a)$ soit minimal. $E(a)$ est en fait l'équation d'une parabole: elle a la forme
$$
E(a)=B\cdot a^2-2C\cdot a + D,
$$
avec $B=\sum_{i=1}^Nx_i^2$, $C=\sum_{i=1}^Nx_iy_i$, et $D=\sum_{i=1}^N y_i^2$. $B$ étant forcément positif cette parabole sera **convexe** et donc
nous sommes assurés qu'il existe un minimum pour $E(a)$. Une façon de déterminer $a$, tel que $E(a)$ est minimal est d'utiliser la dérivée.
On a l'équation $E'(a)=0$ à résoudre:
\begin{align}
E'(a)&=0,\nonumber\\
2\cdot B\cdot a-2\cdot C&=0,\nonumber\\
a &= \frac{C}{B}=\frac{\sum_{i=1}^Nx_iy_i}{\sum_{i=1}^Nx_i^2}.
\end{align}

---

Exemple +.#

Soient les 4 points $(0, 0.1)$, $(1, 0.3)$, $(2, 0.3)$ et $(3, 0.4)$. La fonction d'erreur $E(a)$ s'écrit
$$
E(a)=14\cdot a^2-4.2\cdot a + 0.35.
$$
On peut la représenter comme sur la @fig:e_a et on constate qu'elle possède un minimum proche de $a=0$.

![La fonction $E(a)=14a^2-4.2a+0.35$ pour $a\in[-1,1]$. On voit bien qu'elle possède un minimum proche de $a=0$.](figs/e_a.svg){#fig:e_a width=70%}

En résolvant $E'(a)=0$, on obtient $a=4.2/24=0.15$. On a que l'équation de la droite passant par $(0,0)$ et au plus proche de nos 4 points est
$$
y(x)=0.15\cdot x.
$$
On peut observer le résultat de la régression sur la @fig:regression_ex, où on voit les 4 points (en noir), ainsi que la droite obtenue (en trait bleu).

![Les 4 points $(0, 0.1)$, $(1, 0.3)$, $(2, 0.3)$ et $(3, 0.4)$ (en noir) et la droite obtenue par régression linéaire
(en bleu).](figs/regression_ex.svg){#fig:regression_ex width=70%}

---

La régression linéaire est un problème **d'optimisation continu**. On voit que contrairement au problème du voyageur du commerce,
l'ensemble des solutions est $a\in\real$ et non une suite discrète de villes. Ce genre de problème, bien que possédant un espace de recherche infini,
est bien souvent plus simple à résoudre, car il possède un cadre théorique mieux défini.

Pour le résoudre, nous avons commencé, comme pour le problème du voyageur du commerce par faire un modèle mathématique.
Nous avons construit une fonction à minimiser, $E(a)$, et ajouté une contraite, la forme de $y(x)$. Puis, il a suffit de trouver le minimum de $E(a)$
sous la contrainte et le tour était joué.

## L'optimisation mathématique

Suite à ces deux exemples, nous allons essayer de définir de façon assez théorique comment formuler mathématiquement un problème d'optimisation.
Il existe comme on vient de le voir deux types disctincts de problèmes d'optimisation:

1. L'optimisation continue.
2. L'optimisation discrète (souvent appelée optimisation combinatoire).

### L'optimisation continue

L'optimisation continue ou *programme mathématique continu* est un programme d'optimisation soumis à certaines contraintes.
On peut l'exprimer de la façon suivante.

Soit $f:\real^n\rightarrow\real$ une fonction objectif (ou fontion de coût), on cherche $\vec x_0\in\real^n$, tel que $f(\vec x_0)\leq f(\vec x)$ pour $\vec x$ certaines conditions: les contraintes qui sont en général des égalités strictes ou des inégalités qui peuvent s'exprimer de la façon suivante.
Soient $m$ fonctions $g_i:\real^n\rightarrow\real$
\begin{align}
&g_i(\vec x)\leq 0,\quad i=1,...,m.
\end{align}
Si $m=0$ on a à faire à un problème d'optimisation sans contraintes. On peut résumer tout cela comme
\begin{align*}
&\min_{\vec x\in\real^n}f(\vec x),\\
&g_i(\vec x)\leq 0,\quad i=1,...,m,\\
&\mbox{pour }m\geq 0.
\end{align*}
Les contraintes limites l'espace des solutions et forment un sous-ensemble, noté $A$, de $\real^n$ ($A\subseteq\real^n$).

Une des difficultés pour déterminer le minimum d'une fonction coût est l'existence de plusieurs minima locaux.
Un **minimum local**, $\vec x^\ast\in A$, est tel que pour une région proche de $\vec x^\ast$, on a que $f(\vec x)\geq f(\vec x^\ast)$. Une fonction particulièrement problématique où il est
très difficile de trouver le minimum global est une fonction de Ackley. En une dimension, elle est de la forme (voir la @fig:ackley)
$$
f(x)=-20e^{-0.2*\sqrt{0.5x^2}}-e^{0.5(\cos(2\pi x))}+e+20.
$$

![La fonction d'Ackley en une dimension. Il est particulièrement compliqué de trouver le minimum global en raison de l'existence d'une multitude de minima locaux.](figs/ackley.svg){#fig:ackley width=70%}

On constate la présence d'un grand nombre de minima locaux qui rendent la recherche du minimum global (se trouvant en $x=0$) particulièrement compliqué à déterminer.

L'optimisation continue est très communément utilisée en apprentissage automatique (machine learning), en particulier pour
optimiser les poids des réseaux de neurones.

### Optimisation discrète

L'optimisation discrète, ou *programme mathématique discret*, est très similaire dans la définition mathématique à l'optimisation continue. L'ensemble $\real^n$ est remplcé par un ensemble discret, $\integer^n$. On peut le formuler de la façon suivante. Soit $f:\integer^n\rightarrow \real$ on cherche à résoudre
\begin{align*}
&\min_{\vec x\in\integer^n}f(\vec x),\\
&g_i(\vec x)\leq 0,\quad i=1,...,m,\\
&\mbox{pour }m\geq 0\mbox{ et }\vec x\in\integer^n.
\end{align*}
On voit donc qu'on a une "contrainte" supplémentaire qui est de supposer que les éléments sont tous discrets.

Il existe un grand nombre d'application quotidiennes
de l'optimisation discrète: le choix du moyen optimal de livraison d'un paquet, le meilleur réseau aérien, dans la logistique, le design de microprocesseurs, ...

L'optimisation discrète est une branche des mathématiques très intéressante. Elle permet de poser des problèmes relativements simples, dont les solutions brute force sont très "simples" conceptuellement. En revanche les solutions "réalisables" dans un temps raisonnables (polynomial) sont souvent très difficiles à obtenir comme on l'a vu avec le problème du voyageur de commerce. Le problème se formule de façon simple, il existe une soltution très simple conceptuellement au problème (l'essai de toutes les possibilités), mais qui n'est pas réalisable en pratique, car déjà à 25 villes il faudrait des milliards d'années pour le résoudre.

Une heuristique possible serait de penser les problèmes d'optimisation discrète comme un cas particulier des problèmes continus.
En résumé on *relaxe* (on retire la condition que $\vec x\in\natural^n$) le problème discret en le rendant continu,
on trouve la solution et finalement on tronque la solution pour la rendre entière.

En procédant de la sorte, il est possible d'approximer de trouver une solution approchée de relativement bonne qualité, mais
on est pas du tout certains de trouver la solution optimale.

On peut s'en convaincre avec le problème suivant. On cherche
\begin{align}
&\max 10x_1+11x_2,\\
&10x_1+12x_2\leq 59,\\
&x_1,x_2\in \real\mbox{ ou } x_1,x_2\in \natural.
\end{align}

Dans le cas discret la solution de ce problème est $x_1=1$, $x_2=4$
\begin{align}
&10x_1+11x_2=54,\\
&10x_1+12x_2=58\leq 59.
\end{align}

1. L'heuristique de la relaxation au cas continu, nous donne $x_1^\ast=5.9$, $x_2^\ast=0$

    \begin{align}
    &10x_1+11x_2=59,\\
    &10x_1+12x_2=59\leq 59.
    \end{align}
    On aurait donc après avoir tronqué $x_1^\ast$ et $x_2^\ast$, la solution est donc $x_1=5$, $x_2=0$, on obtient
    $$
    10x_1+11x_2=50,
    $$
    ce qui est assez différent de $x_1=1$ et $x_4=4$, où 10x_1+11x_2=58 (mais c'est pas non plus la pire solution possible).

2. Il est possible d'appliquer une heuristique un peu meilleure en général. Le problème qui nous intéresse ici peut se formuler de la façon suivante en général:

    \begin{align}
    &\max m_1x_1+m_2x_2,\\
    &v_1x_1+v_2x_2\leq V.
    \end{align}
    On commence par calculer "la densité" $\rho_i=m_i/v_i$ de chaque indice $i$. Ici on a $\rho_1=1$ et $\rho_2=11/12$.
    On classe les $\rho_i$ par ordre décroissant, et on prend le maximum possible de chacun.
    On commence donc par prendre autant de $i=1$ que possible comme il a la plus grande densité.
    Cela nous donne $x_1=5$ et il ne reste plus la place pour des $x_2$ telle que la contrainte est satisfaire et on a terminé.
    On a donc $x_1=5$ et $x_2=0$, ce qui est le même résultat que la relaxation continue, mais ce n'est pas toujours le cas.
    Cette façon de procéder est au moins aussi bonne que la relaxation continue en général (elle peut donc être meilleure).

## Programmation linéaire

Contrairement à ce qu'on pourrait croire, la *programmation linéaire* n'a pas forcément grand chose à avoir avec l'informatique. Il s'agit d'un sous ensemble des problèmes d'optimisation mathématique. Si la fonction objective et les fonctions de contraintes sont toutes **linéaires**, alors le problème en lui est même est dit linéaire. La façon la plus générale d'écrire un programme linéaire discret, est de faire un "mélange" (le programme est dit *mixte*) entre un programme discret et un programme continu
\begin{align}
\min_{\vec x_1\in\real^n,\vec x_2\in\integer^n} \vec c_1^T\cdot \vec x_1+ \vec c_2^T\cdot \vec x_2,\\
\mat{A}_1\cdot \vec x_1+\mat{A}_2\cdot \vec x_2\leq b,
\end{align}
où $\vec c_1$ et $\vec c_2$ sont des vecteurs, et $\mat{A}_1$ et $\mat{A}_2$ des matrices des parties continues, $\vec x_1$ et discrfètes, $\vec x_2$.

Les programmes linéaires continus ont tous des solutions polynomiales, en revanche dans le cas discret cela n'est pas le cas comme on a pu le voir avec le cas du voyageur du commerce.

Il existe bien évidemment des programmes non-linéaires (la fonction objectif, ou les fonctions de contraintes ne sont pas linéaires) mais leur résolution est en général beaucoup plus compliquée, car nous ne sommes pas assurés de l'existence d'un minimum global (il existe également des minima locaux).

### Pogrammation linéaire continue: l'algorithme du simplexe

Une des méthodes les plus répandues de résolution de problèmes d'optimisation continue est la méthode du simplexe linéaire.
Bien que cette méthode soit sous-optimale (elle est a en général un coût exponentiel en fonction du nombre de dimensions)
elle reste très utilisée, car elle est relativement intuitive et simple à implémenter.

Nous commençons par définir un problème de programmation linéaire continu de façon générale comme
\begin{align}
&\min_{\vec x\in\real^n} {\vec c}^T\cdot \vec x,\\
&\mat{A}\cdot \vec x\leq b,\\
&\vec x \geq \vec 0,\quad \vec x\in\real^n.
\end{align}
Si on suppose que $\mat{A}$ représente une application surjective, on peut représenter
$\mat{A}$ comme un polyèdre fermé.

---

Exemple (Polygone) +.#

Soit le problème de programmation linéaire suivant
\begin{align}
&\max 10x_1+11x_2,\\
&10x_1+12x_2\leq 59,\\
&x_1\geq 0,\quad x_2\geq 0.
\end{align}
On peut faire correspondre la formulation générale ci-dessus à
\begin{align}
\vec x = \vectwo{x_1}{x_2},\quad \vec c=\vectwo{10}{11},\\
\mat{A}=\left(10\ 12\right),\quad \vec b=(59).
\end{align}
Les conditions sur $\vec x$
\begin{align*}
&x_2\leq \frac{59-10 x_2}{12},\\
&x_1\geq 0,\quad x_2\geq 0,
\end{align*}
peuvent se représenter graphiquement par un polygone (nous sommes en deux dimensions), comme représenté sur la @fig:polygon.

![Les conditions $x_2\leq\frac{59-10 x_2}{12}$ (noir), $x_1\geq 0$ (rouge), $x_2\geq 0$ (bleu) représentées graphiquement. Tout ce qui est à l'intérieur du triangle satisfait ces trois conditions simultanément.](figs/polygon.svg){#fig:polygon width=70%}

---

On peut montrer que si toutes les conditions ci-dessus sont satisfaites, alors la solution optimale se trouve sur un des sommets du polyèdre.
On résout ce genre de problème en parcourant donc tous les sommets en suivants les arrêtes du polyèdre.
Comme vous avez déjà vu cet algorithme dans le cours d'algorithmique avancée, on ne va pas s'attarder dessus,
mais plutôt parler de modélisation d'un problème d'optimisation. En d'autres termes, comment en partant
d'un problème d'optimisation formulé en phrases on écrit un problème mathématique.

## Modélisation

### Exemple: Le problème du sac à dos

Le problème du sac à dos est un problème classique en optimisation combinatoire. Il n'existe pas de méthode exacte qui
permette de résoudre ce problème en temps polynomial (pour ceux que ça intéresse, je pense que résdoudre ce problème de façon générale
vaut beaucoup de gloire et d'argent).
Il s'énonce de la façon suivante. Soit un ensemble de $n$ objets de poids $p_i$ et de valeur $v_i$. Combien de chaque objet
pouvons nous mettre dans le sac à dos, tel que la valeur totale des objets dans le sac à dos soit maximisée,
et que le poids total n'excède pas un poids maximal $P$. Nous simplifions encore le problème en spécifiant
qu'il peut y avoir au plus un objet de chaque dans le sac.

Nous essayons de maximiser la valeur du sac. Nous représentons la présence ou non de l'objet $i$, par une variable $x_i$:
$x_i=1$ si l'objet est présent $x_i=0$ si l'objet est absent. On cherche donc un vecteur $\vec x=(x_1,x_2,...,x_n)^T$, avec $x_i\in\{0,1\}$.
La fonction objectif à maximiser est par conséquent
$$
f(\vec x)=\sum_{i=1}^n x_i\cdot v_i.
$$
Le poids du sac s'exprime lui avec $p(\vec x)$ comme
$$
p(\vec x)=\sum_{i=1}^n x_i\cdot p_i.
$$
On formule donc mathématiquement notre problème d'optimisation comme
\begin{align}
&\max \sum_{i=1}^n x_i\cdot v_i,\\
&\sum_{i=1}^n x_i\cdot p_i\leq P,\\
&x_i\in\{0,1\}.
\end{align}

### Généralités

L'exercice de transposer un problème d'optimisation en formulation mathématique est appelée **modélisation**. L'exercice en lui même est intéressant, et nous allons en voir quelques exemples que vous devrez proposer et résoudre par vous mêmes. Mais avant d'en arriver là, on peut écrire un certain nombre de règles.

Soient deux événement, $A$ et $B$, et les variables de décision
binaires, $x_A,\ x_B\in\{0,1\}$

- Si les deux événements ne peuvent pas se produire en même temps, on peut écrire
$$
x_A+x_B\leq 1.
$$
On a 4 possibilités:
    1. $A$ se produit, $x_A=1$, et $B$ se produit $x_B=1$. On a $x_A+x_B=2$ (c'est plus grand que $1$ nous sommes sauvés).
    2. $A$ se produit, $x_A=1$, et $B$ ne se produit pas $x_B=0$. On a $x_A+x_B=1$ (c'est égal à $1$ nous sommes satisfaisons la contrainte).
    3. $A$ ne se se produit pas, $x_A=0$, et $B$ se produit $x_B=1$. On a $x_A+x_B=1$  (c'est égal à $1$ nous sommes satisfaisons la contrainte).
    4. $A$ ne se se produit pas, $x_A=0$, et $B$ ne se produit pas $x_B=0$. On a $x_A+x_B=0$  (c'est égal à $0$ nous sommes satisfaisons la contrainte).

- Si au moins un des deux événements doit se produit
$$
x_A+x_B\geq 1.
$$
On a 4 possibilités:
    1. $A$ se produit, $x_A=1$, et $B$ se produit $x_B=1$. On a $x_A+x_B=2$ (c'est plus grand que $1$ satisfaisons la contrainte).
    2. $A$ se produit, $x_A=1$, et $B$ ne se produit pas $x_B=0$. On a $x_A+x_B=1$ (c'est égal à $1$ nous satisfaisons la contrainte).
    3. $A$ ne se se produit pas, $x_A=0$, et $B$ se produit $x_B=1$. On a $x_A+x_B=1$  (c'est égal à $1$ nous satisfaisons la contrainte).
    4. $A$ ne se se produit pas, $x_A=0$, et $B$ ne se produit pas $x_B=0$. On a $x_A+x_B=0$  (c'est plus petit que $1$, nous sommes sauvés).

- Si $A$ se produit, alors $B$ doit se produire
$$
x_A\leq x_B.
$$
Si $A$ s'est produit, nous avons $x_A=1$, cela signifie que $x_B$ doit se produire également et donc $1\leq x_B$, $x_B$ doit forcément valoir $x_B=1$ sinon la contrainte n'est pas satisfaite. De façon similaire, si $B$ ne se produit pas ($x_B=0$), alors $A$ ne doit pas se produire, car $x_A\leq 0$. On a donc que $x_A=0$ obligatoirement.
- Soit $C$ un troisième événement et $x_C$ sa variable de décision binaire. Si $A$ se produit, alors $B$ ou/et $C$ doivent se produire:
$$
x_A\leq x_B+x_C.
$$

Souvent, il est utile de rajouter des variables même si cela complexifie *a priori* le problème. Cela permet néanmoins d'être plus expressif.

- Soient un nombre $N$ de fonctions linéaires $c_i x$,
$i=1, ..., N$. Si l'objectif est de maximiser la valeur minimale prise par cette ensemble de fonction, on peut rajouter une variable $y$ et l'ensemble de contraintes
$$
y\leq c_i x.
$$
La fonction objectif est elle-même modifiée et devient $\max y$.

- Soient $N$ contraintes
\begin{align}
&c_1 x\leq b_1,\\
&c_2 x\leq b_2,\\
&\cdots,\nonumber\\
&c_N x\leq b_N.
\end{align}
Si nous souhaitons en imposer un sous ensemble, $K$ en même temps. Nous pouvons définir une variable $M_i$ qui satisfera la $i$-ème contrainte
quelle que soit $x$
$$
c_i x\leq b_i+M_i,\quad\forall x.
$$
On définit également $y_i$, une variable binaire, pour chaque contrainte. Nous pouvons dès-lors réécrire les contraintes de la façon suivante
\begin{align}
&c_1 x\leq b_1+M_1y_1,\\
&c_2 x\leq b_2+M_2y_2,\\
&\cdots,\nonumber\\
&c_N x\leq b_N+M_Ny_N,\\
&\sum_{i=1}^N y_i = N-K.
\end{align}

---

Question +.#

Que se passe-t-il si $N=2$ et $K=1$?

---

Ci-dessous vous trouverez trois problèmes différents qu'il faudra modéliser. 

---

Problème (Fabrication) +.#

Une entreprise produit des pizzas de luxe (LU) et "low-cost" (LC).
Elles sont toutes les deux fabriquées à partir des mêmes produits,
pâte, fromage, et jambon,
mais dans des proportions différentes qui sont résumées dans le tableau ci-dessous

+----------+--------+-------+
|          | L      | LC    |
+==========+========+=======+
| Pâte     | 1      | 3     |
+----------+--------+-------+
| Fromage  | 2      | 1     |
+----------+--------+-------+
| Jambon   | 2      | 1     |
+----------+--------+-------+

Chacune de ces matières premières est disponible en quantité limitée:
$1200\mathrm{kg}$ de pâte, $600\mathrm{kg}$ de fromage, et $400\mathrm{kg}$ de jambom. La pizza de luxe rapporte $10$ francs par kilo et la pizza low-cost rapporte $6$ francs par kilo. Modélisr le problème du calcul de nombre de kilos de pizzas à produire de chaque type pour maximiser le rendement.

Modéliser le problème:

1. Déterminer les variables du problème.
2. Écrire la fonction objectif.
3. Ecrire les contraintes.

Le problème se modélise de la façon suivante:

1. Les variables sont $x_i$ les quantités en kilos de pizzas du type $LU$ et $LC$ (on a donc $x_{LU}$, et $x_{LC}$).
2. La fonction objectif à maximiser est le rendement des pizzas:
$$
\max_{\vec x} f(\vec x)=\max_{\vec x}10\cdot x_{LU}+6\cdot x_{LC}.
$$
3. Les contraintes sont données par:
\begin{align}
1\cdot x_{LU} + 3\cdot x_{LC}\leq 1200,\\
2\cdot x_{LU} + 1\cdot x_{LC}\leq 600,\\
2\cdot x_{LU} + 1\cdot x_{LC}\leq 400,\\
x_{LU}\geq 0,\quad x_{LC}\geq 0.
\end{align}

---

---

Problème (Distribution d'électricité) +.#

Nous souhaitons alimenter en électricité trois villes, Genève, Bâle, et Zürich.
Nous avons à notre disposition trois centrales (celle de la Grande Dixence, de Leibstadt et
de Mühleberg). Nous souhaitons à partir du tableau ci-dessous, représentant les coût de production
et d'acheminement de l'électricité des centrales aux villes, ainsi que leur demande et puissance
respectives en GWh. Déterminer fournir en électricitié toutes les villes à moindre coût.
Modéliser le problème d'optimisation des coûts globaux d'élctricité (tous les chiffres donnés ici
sont choisis au hasard).

+----------------+--------+--------+--------+-----------------+
|                | Bâle   | Genève | Zürich | Puissance (GWh) |
+================+========+========+========+=================+
| Grande Dixence | 3      | 5      | 8      | 30              |
+----------------+--------+--------+--------+-----------------+
| Leibstadt      | 8      | 11     | 6      | 50              |
+----------------+--------+--------+--------+-----------------+
| Mühleberg      | 11     | 15     | 12     | 40              |
+----------------+--------+--------+--------+-----------------+
| Demande (GWh)  | 20     | 25     | 40     |                 |
+----------------+--------+--------+--------+-----------------+

Modéliser le problème:

1. Déterminer les variables du problème.
2. Écrire la fonction objectif.
3. Ecrire les contraintes.

Le problème se modélise de la façon suivante:

1. Les variables du probèmes sont $x_{ij}$, le nombre de GWh acheminés de la centrale $i$ à la ville $j$.
On peut par exemple numéroter la Grande Dixence, Leibstadt et Mühleberg, 1, 2, et 3 respectivement et
Bâle, Genève et Zûrich, 1, 2, et 3 respectivement.
2. Le coût total de production et d'acheminement de l'électricité et la fonction objectif $f(\vec x)$ qui s'écrit
\begin{align}
f(\vec x) = &3x_{11} + 5x_{12} + 8x_{13} +\nonumber\\
            &8x_{21} + 11x_{22} + 6x_{23} +\nonumber\\
            &11x_{31} + 15x_{32} + 12x_{33}.
\end{align}
3. Il y a trois sortes de contraintes.

    a. Les la quantité d'électricité acheminée doit être positive
    $$
    x_{ij}\geq 0,\quad\forall{i,j}.
    $$
    b. La production maximale de chaque centrale ne doit pas être dépassée
    \begin{align}
    &x_{11}+x_{12}+x_{13}\leq 30,\\
    &x_{21}+x_{22}+x_{23}\leq 50,\\
    &x_{31}+x_{32}+x_{33}\leq 40.
    \end{align}
    c. Chaque ville doit au moins recevoir la quantité d'électricité demandée
    \begin{align}
    &x_{11}+x_{21}+x_{31}\geq 20,\\
    &x_{12}+x_{22}+x_{32}\geq 25,\\
    &x_{13}+x_{23}+x_{33}\geq 40.
    \end{align}

---

---

Problème (Transport) +.#

Le graphe de la @fig:graph_transport représente le réseau de transport d'un jouet de deux usines $U$, et $V$,
à deux clients $C$, et $D$ en passant par les entrepôts $E$ et $F$. Sur les arrêtes du graphe se
situent les coûts de transport par jouet transporté.

![Graphe des chemins possibles entre les usines $U$, $V$ et les clients $C$, $D$, en passant par les entrepôts $E$, $F$. Les arrêtes des graphes représentent les coûts de transport entre chque noeud.](figs/graph_transport.svg){#fig:graph_transport width=100%}

A la sortie de l'usine $A$, il y a une quantité de $90$ jouets disponibles et une quantité de $75$
jouets à la sortie de l'usine $B$. Les clients $C$ et $D$ ont une demande de respectivement
$70$ et $80$. Les entrepôts ne sont qu'un point de transit.

Modéliser le problème d'optimisation permettant de déterminer le planning de transport optimal
qui permet de satisfaire les demandes des clients pour un coût aussi faible que possible.

1. Déterminer les variables du problème.
2. Écrire la fonction objectif.
3. Ecrire les contraintes.

Le problème se modélise de la façon suivante:

1. Les variables sont les quantités transportées entre les usines et entrepôts,
$x_{UE}$, $x_{UF}$, $x_{VE}$, $x_{VF}$, et les quantités transportées entre les
entrepôts et les clients $x_{EC}$, $x_{ED}$, $x_{FC}$, $x_{FD}$.
2. On cherhce à minimiser la fonction de coût de transport $f(\vec x)$
$$
f(x)=x_{UE} + 2x_{UF} + 3x_{VE} + 4x_{VF} + 5x_{EC} + 8x_{ED}
     + 7x_{FC} + 6x_{FD}.
$$
3. Il y a 4 sortes de contraintes.

    a. Les quantités de jouets sont des entiers positifs
    $$
    x_{ij}\in\natural.
    $$
    b. Ce qui part des usines doit être plus petit que le stock local
    \begin{align}
    x_{UE}+x_{UF}&\leq 90,\\
    x_{VE}+x_{VF}&\leq 75.
    \end{align}
    c. Ce qui arrive à chaque entrepôt doit partir de chaque entrepôt
    \begin{align}
    x_{UE}+x_{VE}&=x_{EC}+x_{ED},\\
    x_{UF}+x_{VF}&=x_{FC}+x_{FD}.
    \end{align}
    d. Les clients doivent recevoir la quantité demandée
    \begin{align}
    x_{EC}+x_{FC}&= 70,\\
    x_{ED}+x_{FD}&= 80.
    \end{align}

---

## Optimisation continue

Dans cette section, nous allons considérer des problèmes purement continus.
Nous allons dans un premier temps considérer une fonction opbjectif, $f$,
$$
f:D\rightarrow\real,\quad D\subseteq \real,
$$
dont nous allons chercher le minimum (pour autant qu'il existe). Nous allons supposer que
$f$ est une fonction continue et dérivable.

### Minimum local/global

Comme vous le savez, le minimum (ou le maximum) d'une fonction, se situe à un endroit où sa dérivée est nulle.
On recherche donc, $x$, tel que
$$
f'(x)=0.
$$
Mais cette contrainte sur $f'(x)$ n'est pas suffisante pour garantir de trouver un minimum.
En effet, si $f'(x)=0$, peut également vouloir dire qu'on se trouve sur un point d'inflexion
ou sur un maximum.
On peut assez facilement, discriminer ces deux cas, en considérant la deuxième dérivée de $f$.
En effet, nous avons à faire à un minimum seulement si
$$
f''(x)>0.
$$
Les cas où $f''(x)=0$ est un point d'inflexion et $f''(x)<0$ est un maximum.

Un autre problème beaucoup plus compliqué à résoudre est de déterminer un minimum **global**.
En effet, comme pour la fonction de Ackley (voir la @fig:ackley), une fonction peut posséder un grand nombre de minimam **locaux** (où
$f'(x)=0$ et $f''(x)>0$) mais qui n'est pas un mimumum global.

Mathématiquement un *minimum local* se définit comme $x^\ast$ tel qu'il existe $\delta>0$ et que $f(x^\ast)\leq f(x)$, pour
$x\in[x^\ast-\delta,x^\ast+delta]$. Un *minimum global* est un $x^\ast$ tel que $\forall x\in D$, $f(x^\ast)\leq f(x)$.

En fait, il n'existe pas de méthode pour déterminer un minimum global, pour n'importe quelle fonction.
Nous somme assurés de le trouver, uniquement si $f$ est une fonction convexe partout ($f''(x)>0 \ \forall x$).

## Algorithme de recherche d'un zéro d'une fonction

Comme nous venons de le voir, lors de la recherche d'un minimum, il est nécessaire de trouver le point $x^\ast$
où $f'(x^\ast)=0$. Le problème est donc de déterminer les zéros de la fonction $f'(x)$. Pour avoir un maximum de généralité,
nous allons considérer une fonction $g(x)$ et chercher ses zéros, soit
$$
\{x\real|g(x)=0\}.
$$
Dans des cas simples (des fonctions polynomiales de degré 2 ou 3, ou des fonctions inversibles) on peut trouver
analytiquement les zéros. En revanche, pour des fonctions plus complexes, ou "implicites" (on ne peut pas mettre
l'équation $g(x)=0$ sous la forme $x=...$) la détermination des zéros est beaucoup plus difficile et nécessite l'utilisation
de **méthodes itératives**. Nous allons en voir quelques unes.

### Méthode de la bissection

![Illustration de la méthode de la bissection. Source: Wikipedia
<https://bit.ly/2RcljBW>.](figs/Bisection_method.svg){#fig:bissection_method width=50%}

Afin de déterminer le zéro d'une fonction, une des méthodes les plus simple est la méthode de la bissection.
Il s'agit de choisir deux points, $a_1$ et $b_1$, $b_1>a_1$, tels que le signe de $g(a_1)$ et $g(b_1)$ est différent.
Si cela est le cas, nous aommes assurés de l'existence d'au moins un zéro si la fonction $g(x)$ est continue
(en vertu du théorème de la valeur intermédiaire). Ensuite, nous allons calculer la valeur se situant "au milieu"
entre $a_1$ et $b_1$
$$
c_1=\frac{b_1+a_1}{2}.
$$
Puis, nous évaluons $g(c_1)$ et si ce n'est pas un zéro, étudions son signe. Si le signe $g(c_1)$ est différent de celui de $g(a_1)$, nous remplaçons
$b_1$ par $c_1$ et recommençons. Si le signe de $g(c_1)$ est différent de celui de $g(b_1)$, nous remplaçons $a_1$ par $c_1$.
Nous itérons cette méthode jusqu'à ce que nous ayons atteint une valeur "siffisamment proche" (nous vons une précision acceptable pour nous)
de zéro. Une façon d'exprimer "proche" est de considérer la taille de l'intervalle $b_1-a_1$ et de le comparer avec une précision $\varepsilon>0$ que nous
aurons choisie
$$
b_1-a_1<\varepsilon.
$$

Au pire des cas, cette méthode nous rapproche de $(b_1+a_1)/2$ du zéro à chaque itération. Après $n$ itération, nous somme donc à une
distance maximale du zéro de $(b_1+a_1)/2^n$. On dit que cette méthode est d'ordre $1$ (on divise l'intervalle de recherche par 2 et la précision par 2
à chaque itération).

---

Exercice (Racice de polynôme) +.#

Déterminer la racine du polynôme $x^4+x^3+x^2-1$ avec $a_1=0.5$ et $b_1=1$ (faire au maximum 6 itérations).

---

### Méthode de la fausse position (*regula falsi*)

![Illustration de la méthode de la fausse position. Source: Wikipedia
<https://bit.ly/2FeHdhm>.](figs/False_position_method.svg){#fig:false_position_method width=50%}

Une méthode un peu plus avancée est la méthode de la fausse position (voir la @fig:false_position_method). Dans cette méthode qui est relativement similaire à celle de la bissection,
mais au lieu de diviser l'intervalle en deux parts égales à chaque itération on va choisir les point $c$, comme étant le point
où la droite reliant $g(a_1)$ et $g(b_1)$ coupe l'axe horizontal (le zéro de la droite entre $g(a_1)$ et $g(b_1)$). Le reste de l'algorithme reste exactement le même.
On choisit deux points, $a_1$ et $b_1$, où le signe de $f$ est différent, puis ont construit la droite passant par $g(a_1)$ et $g(b_1)$
$$
y=\frac{g(b_1)-g(a_1)}{b_1-a_1}(x-a_1) + g(a_1).
$$
On cherche le point, $c$, où $y(c)=0$
$$
\frac{g(b_1)-g(a_1)}{b_1-a_1}(c-a_1) + g(a_1)=0.
$$
Cette équation s'inverse aisément et on obtient
$$
c_1=a_1-\frac{b_1-a_1}{g(b_1)-g(a_1)}g(a_1).
$$
Puis, comme pour la méthode de la bissection, on compare les signes de $g(c_1)$ avec $g(a_1)$ et $g(b_1)$ et on remplace $a_1$ ou $b_1$ par $c_1$
si $g(c_1)$ a un signe différent de $g(b_1)$ ou $g(a_1)$ respectivement.

Il est important de noter que si la fonction est continue, et que $a_1$ et $b_1$ sont choisis tels que $g(a_1)$ et $g(b_1)$ sont de signes opposés,
alors cette méthode convergera **toujours**.

La méthode de la fausse position est plus efficace que la méthode de la bissection, elle est superlinéaire (d'ordre plus grand que un).

---

Exercice +.#

Déterminer le zéro positif de la fonction
$$
x^2-25=0,
$$
à l'aide de la méthode de la fausse position.

---

### Méthode de la sécante

![Illustration de la méthode de la sécante. En partant des points $x_0$ et $x_1$, on s'approche du zéro
en calculant le point $x_2$. Source: Wikipedia
<https://bit.ly/2FcAXpA>.](figs/Secant_method.svg){#fig:secant_method width=50%}

La méthode de la sécante (voir la @fig:secant_method) est très similaire à la méthode de la fausse position. La seule différence se situe dans la dernière étape de l'algorithme.
Plutôt que choisir de remplacer $a_1$ ou $b_1$ par $c_1$, on remplace toujours la dernière valeur calculée.
Ainsi après avoir choisi $a < b$, avec $g(a)$ et $g(b)$ avec des signes différents, on calcule
une suite de $x_i$, avec $x_0=a$, $x_1=b$, tels que
$$
x_{i+1}=x_{i-1}-\frac{x_i-x_{i-1}}{g(x_i)-g(x_{i-1})}g(x_{i-1}), \quad  i\geq 2.
$$

La méthode de la sécante ne garantit pas la convergence, contrairement à la méthode de la bissection et de la fausse position.
En revanche elle est plus efficace, lorsque qu'elle converge, que ces deux méthodes.

---

Exercice +.#

Déterminer le zéro positif de la fonction
$$
x^2-25=0,
$$
à l'aide de la méthode de la sécante.

---

### Méthode de Newton (ou *Newton-Raphson")

La méthode de Newton est également une méthode itérative, qui nécessite que la fonction $g(x)$ soit non seulement continue mais également dérivable.
Revenons sur la méthode de la sécante. Il s'agissait de choisir deux points, $a < b$, et de déterminer la droite, $y(x)$, passant par $g(a)$ et $g(b)$,
\begin{equation*}
y=\frac{g(b)-g(a)}{b-a}(x-a) + g(a).
\end{equation*}
Il se trouve que $g(b)-g(a)/(b-a)$ n'est autre qu'une approximation avec une formule de différences finies
de la dérivée de $g$ et $a$, $g'(a)$. Si la fonction $f$ est dérivable, on peut simplement remplacer ce terme par $g'(a)$
et on obtient
$$
y=g'(a)(x-a) + g(a).
$$
Puis on détermine $c$, tel que $y(c)=0$
$$
0=g'(a)(c-a) + g(a),
$$
et on obtient
$$
c=a - \frac{g(a)}{g'(a)}.
$$

On peut donc généraliser l'algorithme. En partant d'un point $x_0=a$, on construit la suite
$$
x_{i+1}=x_n-\frac{g(x_i)}{g'(x_i)}, \ i\geq 0.
$$
On s'arrête lorsque le zéro est déterminé avec une précision suffisante, ou que la variation entre deux itérations successives est assez petite. Ce qui revient à choisir un $\varepsilon>0$, tel que
$$
|g(x_n)| < \varepsilon,\quad |x_n-x_{n-1}| < \varepsilon.
$$

Lorsque qu'elle converge la mtéhode de Newton est la plus efficace de toutes celles que nous avons vues. On dit qu'elle est d'ordre $2$.
En revanche les contraintes pour sa convergence sont plus strictes que pour les méthodes vues précédemment.

---

Remarques (non-convergence ou convergence lente) +.#

Il y a un certain nombre de cas où la méthode de Newton ne converge pas.

1. S'il existe un $n$ tel que $g'(x_n)=0$ alors la suite diverge.
2. La suite peut entrer dans un cycle.
3. La dérivée est mal définie proche du zéro (ou sur le zéro).
4. Elle peut converger très lentement si la dérivée de la fonction est nulle sur le zéro.
5. A chaque point de départ ne correspond qu'un zéro. Si la fonction possède plusieurs zéros, il n'y a pas moyen de le savoir avec un seul point de départ. Il faut alors en essayer plusieurs.

---

---

Exercice +.#

Déterminer le zéro de la fonction
$$
x^2-25=0,
$$
à l'aide de la méthode de Newton.

---

### Résumé

A l'aide des méthodes vues ci-dessus, on peut déterminer un zéro d'une fonction (s'il existe).
Ces méthodes sont également utilisables pour calculer le minimum d'une fonction comme nous l'avons discuté plus haut.
Il suffit de remplacer $g(x)$ par $f'(x)$ et le tour est joué.

---

Exercice +.#

Écrire l'algorithme de Newton pour le cas de la minimisation d'une fonction $f(x)$ quelconque, mais continuement dérivable 2 fois.

---

<!-- ## Le cas multivarié

### La descente de gradient

Cette méthode est **la** méthode utilisée dans les réseaux de neurones que vous verrez dans votre cours d'aprrentissage automatique (Machine Learning). -->

# Remerciements

Je voudrais remercier (par ordre alphabétique) les étudiants du cours
qui ont contribué à améliorer ce polycopié. En espérant que cette liste
continuera à s’allonger avec les années. Merci à Messieurs
Heirich et Pirkl.

[^1]: On pourrait, de façon similaire, utiliser la formule de différences finies en avant ou en arrière (ou un mélange des deux).
[^2]: Comme ce polynôme passe par les points $(x_0,y_0)$, $(x_1,y_1)$, ..., $(x_m,y_m)$, il est unique, c'est donc exactement le même que celui exprimé avec les $\{a_i\}_{i=0}^m$.
[^3]: En fait $\frac{f^{(n+1)}(\xi)}{(n+1)!}=\delta^{n+1}[x_0,\dots,n_n]$.
[^4]: Elle satisfait également $\int_a^b \left(s''(x)\right)^2\rightarrow \mathrm{min}$ mais n'entrons pas trop dans les détails.
