import numpy as np
import matplotlib.pyplot as plt
import math

def f(x): 
	return np.sin(x)

def df(x): 
	return np.cos(x)

def app(x, h): 
	return (f(x+h)-f(x)) / h

def comp_error_sqr(xmin, xmax, h):
	x = np.arange(xmin, xmax, h)
	approx = app(x,h)
	exact = df(x)

	error = (approx - exact)**2
	return x, error

def comp_error_moy(error):
	q = error.size
	return np.sum(np.sqrt(error) / q)

def comp_error_tot(error):
	return np.sum(np.sqrt(error))



h = 0.2
xmin = 0.0
xmax = math.pi

x, error = comp_error_sqr(xmin, xmax, h)

plt.figure(1)

plt.subplot(121)
plt.title('Fonction')
plt.plot(x, df(x), 'b-', label="$f'(x)=\cos(x)$")
plt.plot(x, app(x, h), 'rx-', label="approx.")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.plot(x, error, 'b-')
plt.ylabel('Erreur')
plt.xlabel('$x$')
plt.grid(True)
# plt.savefig("err_cos.svg")
plt.show()

h = [0.025, 0.05, 0.1, 0.2, 0.4]
xmin = 0.0
xmax = math.pi

e_tot = np.array([])
e_moy = np.array([])

for dx in h:
	x, error = comp_error_sqr(xmin, xmax, dx)
	e_tot = np.append(e_tot, comp_error_tot(error))
	e_moy = np.append(e_moy, comp_error_moy(error))


plt.figure(2)

plt.plot(h, e_tot, 'b-', label="$E_\mathrm{tot}$")
plt.plot(h, e_moy, 'rx-', label="$E_\mathrm{moy}$")
plt.legend(loc=0)
plt.ylabel('Erreur')
plt.xlabel('$h$')
plt.grid(True)
# plt.savefig("err_deriv_avant.svg")
plt.show()

