import numpy as np
import matplotlib.pyplot as plt
import math

def f(x): 
	return np.sin(x)

def df(x): 
	return np.cos(x)

def app(x, h): 
	return (f(x+h)-f(x)) / h

def app_c(x, h): 
	return (f(x+h/2)-f(x-h/2)) / h

def comp_error_sqr(xmin, xmax, h):
	x = np.arange(xmin, xmax, h)
	approx = app(x,h)
	exact = df(x)

	error = (approx - exact)**2
	return x, error

def comp_error_sqr_c(xmin, xmax, h):
	x = np.arange(xmin, xmax, h)
	approx = app_c(x,h)
	exact = df(x)

	error = (approx - exact)**2
	return x, error

def comp_error_moy(error):
	q = error.size
	return np.sum(np.sqrt(error) / q)

def comp_error_tot(error):
	return np.sum(np.sqrt(error))



h = 0.2
xmin = 0.0
xmax = math.pi

x, error = comp_error_sqr(xmin, xmax, h)
x, error_c = comp_error_sqr_c(xmin, xmax, h)

plt.figure(1)

plt.subplot(121)
# plt.title('Fonction')
plt.plot(x, df(x), 'b-', label="$f'(x)=\cos(x)$")
plt.plot(x, app(x, h), 'rx-', label="avant")
plt.plot(x, app_c(x, h), 'gx-', label="centr.")
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)

plt.subplot(122)
plt.loglog(x, error, 'b-', label="Erreur en avant")
plt.loglog(x, error_c, 'g-', label="Erreur centrée")
plt.ylabel('Erreur')
plt.xlabel('$x$')
plt.legend(loc=0)
plt.grid(True)
plt.savefig("err_cos_ac.svg")
# plt.show()

h = [0.025, 0.05, 0.1, 0.2, 0.4]
xmin = 0.0
xmax = math.pi

e_tot = np.array([])
e_tot_c = np.array([])
e_moy = np.array([])
e_moy_c = np.array([])

for dx in h:
	x, error = comp_error_sqr(xmin, xmax, dx)
	e_tot = np.append(e_tot, comp_error_tot(error))
	e_moy = np.append(e_moy, comp_error_moy(error))

	x, error = comp_error_sqr_c(xmin, xmax, dx)
	e_tot_c = np.append(e_tot_c, comp_error_tot(error))
	e_moy_c = np.append(e_moy_c, comp_error_moy(error))


plt.figure(2)

plt.loglog(h, e_tot, 'b-', label="$E_\mathrm{tot,avant}$")
plt.loglog(h, e_moy, 'rx-', label="$E_\mathrm{moy,avant.}$")
plt.loglog(h, e_tot_c, 'k-', label="$E_\mathrm{tot,centr.}$")
plt.loglog(h, e_moy_c, 'gx-', label="$E_\mathrm{moy,centr.}$")
plt.legend(loc=0)
plt.ylabel('Erreur')
plt.xlabel('$h$')
plt.grid(True)
plt.savefig("err_deriv_avant_centree.svg")
# plt.show()

