import numpy as np
import matplotlib.pyplot as plt
import math

def echant(x0, x1, n, f):
	dx = (x1-x0)/n
	x = np.arange(x0,x1+dx/10,dx)
	return (x,f(x))

def interp_piece(x, x_p1, y, y_p1, dx):
	a = (y_p1 - y) / (x_p1 - x)
	b = - (x*y_p1 - x_p1 * y) / (x_p1-x)

	loc_x = np.arange(x, x_p1, dx)

	return (loc_x, loc_x * a + b)

def interp(x0, x1, n1, n2, f):
	dx = (x1 - x0) / n1
	all_x = np.array([])
	all_y = np.array([])
	for i in range(n1):
		x = x0 + i*dx
		x_p1 = x0 + (i+1)*dx

		loc_dx = (x_p1 - x) / n2
		(loc_x, loc_y) = interp_piece(x, x_p1, f(x), f(x_p1), loc_dx)

		if np.size(loc_x) > n1: # sometimes ranges goes too far
			loc_x = loc_x[0:n1]
			loc_y = loc_y[0:n1]


		all_x = np.append(all_x, loc_x)
		all_y = np.append(all_y, loc_y)
		
	return (all_x, all_y)

res = np.array([])
error = np.array([])
for n in range(1000,10000,100):
	x0 = -5
	x1 = 5
	(x,y) = echant(x0, x1, n, np.sin)

	(x,y) = interp(x0, x1, n, n, np.sin)

	n *= n
	(x_a,y_a) = echant(x0, x1, n, np.sin)

	error = np.append(error, np.sqrt(np.sum((y_a[0:n]-y)**2) / n))
	res = np.append(res, math.sqrt(n))

plt.loglog(res, error, 'ko-', label="Error")
plt.loglog(res, res**(-2), 'g-', label="$N^{-2}$")
plt.xlabel("$N$")
plt.legend(loc=0)
plt.grid(True)
plt.show()

# plt.savefig('error_deg1.svg')

